import sys
from acecurity.crypto import PasswordManager, set_backend, Backend, PQPasswordManager, DataEncryptor, DigitalSigner
from acecurity import Security

set_backend([Backend.cryptography])

# ret: bytes = PasswordManager.hash_password("Test", strength=Security.SUPER_STRONG)
# print("RET", ret)
# print(PasswordManager.verify_password("Test", ret))

# ret: bytes = PQPasswordManager.hash_password("Test", strength=Security.STRONG)
# set_backend([Backend.argon2_cffi])
# print(PQPasswordManager.verify_password("Test", ret))

# ec = DataEncryptor.generate()
# print(ec.get_key())
# crypt: bytes = ec.encrypt_data(b"Test data")
# print(crypt)
# key: bytes = ec.get_key()
# new_ec = ec.from_key(key)
# print(new_ec.decrypt_data(crypt))

sign = DigitalSigner.generate()
print(sign.get_private_key())
signature: bytes = sign.sign_data(b"My data")
print("SIG", signature)
print(sign.verify_signature(b"My data", signature))

sys.exit(0)

from acecurity.tests.test_rand import test_random_generators, test_weighted_functions


test_random_generators()
# test_weighted_functions()

from acecurity.passwords import SecurePasswordGenerator

generator = SecurePasswordGenerator(security="super_strong")
pw_data = generator.passphrase()

print(pw_data["password"])     # Actual password
print(pw_data["extra_info"])   # Generation method metadata

generator.sentence()         # Readable sentence
generator.pattern()          # Format like Aa99##
generator.complex_pattern()  # Pattern + random word mix
generator.complex()          # High entropy, mixed format
generator.mnemonic()         # Easy to remember

pw_data = generator.generate_secure_password(return_worst_case=True)
print(pw_data["worst_case"])  # e.g., "centuries"

from acecurity.crypto.algos import (
    Sym,
    Asym,
    HashAlgorithm,
    KeyDerivationFunction,
)
from acecurity.crypto.exceptions import NotSupportedError
from acecurity.crypto import set_backend, Backend
import os


set_backend()  # Uses Backend.std_lib


# MD5
h = HashAlgorithm.MD5.hash(b"hello world")
print("MD5:", h.hex())
print("Verify:", HashAlgorithm.MD5.verify(b"hello world", h))
print("Std-Verify:", HashAlgorithm.std_verify(b"hello world", h))

# SHA1
h = HashAlgorithm.SHA1.hash(b"hello world")
print("SHA1:", h.hex())
print("Verify:", HashAlgorithm.SHA1.verify(b"hello world", h))
print("Std-Verify:", HashAlgorithm.std_verify(b"hello world", h))

# SHA256
h = HashAlgorithm.SHA2.SHA256.hash(b"hello world")
print("SHA256:", h.hex())
print("Verify:", HashAlgorithm.SHA2.SHA256.verify(b"hello world", h))
print("Std-Verify:", HashAlgorithm.std_verify(b"hello world", h))

# SHA3-256
h = HashAlgorithm.SHA3.SHA256.hash(b"hello world")
print("SHA3-256:", h.hex())
print("Verify:", HashAlgorithm.SHA3.SHA256.verify(b"hello world", h))
print("Std-Verify:", HashAlgorithm.std_verify(b"hello world", h))


# SHAKE128 (8 bytes)
h = HashAlgorithm.SHA3.SHAKE128.hash(b"hello", 8)
print("SHAKE128:", h.hex())
print("Verify:", HashAlgorithm.SHA3.SHAKE128.verify(b"hello", h))
print("Std-Verify:", HashAlgorithm.std_verify(b"hello", h))

# BLAKE2s (8 bytes)
h = HashAlgorithm.BLAKE2.BLAKE2s.hash(b"hello", 8)
print("BLAKE2s:", h.hex())
print("Verify:", HashAlgorithm.BLAKE2.BLAKE2s.verify(b"hello", h))
print("Std-Verify:", HashAlgorithm.std_verify(b"hello", h))


# RIPEMD160
try:
    h = HashAlgorithm.RIPEMD160.hash(b"hello")
    print("RIPEMD160:", h.hex())
    print("Verify:", HashAlgorithm.RIPEMD160.verify(b"hello", h))
    print("Std-Verify:", HashAlgorithm.std_verify(b"hello", h))
except Exception as e:
    print("RIPEMD160 unsupported in std_lib:", e)


password = b"my-password"
salt = os.urandom(16)

print("PBKDF2HMAC:", KeyDerivationFunction.PBKDF2HMAC.derive(password, salt=salt).hex())
print("PBKDF1    :", KeyDerivationFunction.PBKDF1.derive(password, salt=salt, length=16).hex())
print("Scrypt    :", KeyDerivationFunction.Scrypt.derive(password, salt=salt, length=16).hex())
print("HKDF      :", KeyDerivationFunction.HKDF.derive(password, salt=salt).hex())
print("ConcatKDF :", KeyDerivationFunction.ConcatKDF.derive(password, otherinfo=b"my-info").hex())


set_backend(
    [
        Backend.argon2_cffi,         # Required for Argon2; Argon2 is also supported by cryptography but not 100% so we put quantcrypt first
        Backend.cryptography,
        Backend.pycryptodomex,
        Backend.quantcrypt,          # To enable post-quantum cryptography
        Backend.bcrypt,              # Required for BCrypt
        Backend.std_lib,             # Fallback
    ]
)


# Hash a password/message using Argon2
hashed = HashAlgorithm.ARGON2.hash(b"Ha", os.urandom(16))
print("\nArgon2 Hash:", hashed.decode())

# Verify the hash
is_valid = HashAlgorithm.ARGON2.verify(b"Ha", hashed)
print("Argon2 Valid:", is_valid)

try:
    print(
        "Std-Verify",
        HashAlgorithm.std_verify(
            b"Ha", hashed, fallback_algorithm="argon2", text_ids=False
        ),
    )  # Std-Verify can't decode special algos like argon2 or bcrypt
except NotSupportedError:
    print("Std-Verify failed")


# Hash a password with BCrypt
bcrypt_hash = HashAlgorithm.BCRYPT.hash(b"my-secret-password")
print("BCrypt Hash:", bcrypt_hash.decode())

# Verify the password against the hash
is_valid = HashAlgorithm.BCRYPT.verify(b"my-secret-password", bcrypt_hash)
print("BCrypt Valid:", is_valid)


# Derive a key using Argon2 KDF
derived_key = KeyDerivationFunction.ARGON2.derive(b"my-password", salt=os.urandom(16))
print("Argon2 Derived Key:", derived_key.hex())

# Derive a key using BCrypt KDF
bcrypt_key = KeyDerivationFunction.BCRYPT.derive(b"my-password", salt=os.urandom(16))
print("BCrypt Derived Key:", bcrypt_key.hex())


# Recipient generates a keypair
recipient_key = Asym.Cipher.KYBER.keypair.new("kyber1024")

# Extract public key from recipient and share it with the sender
pub_key_bytes = recipient_key.encode_public_key()
# Keys can't be regenerated and try: except: takes more space;
# This can only happen if you do not pass one of the keys when using .decode( ... )
if pub_key_bytes is None:
    raise ValueError("recipient_key has no public key")

# Sender receives the public key and creates a key object with only the public key
sender_key = Asym.Cipher.KYBER.keypair.decode("kyber1024", public_key=pub_key_bytes)
# Sender encapsulates a shared secret for the recipient
ciphertext, shared_secret_sender = sender_key.encapsulate()

# Recipient decapsulates to recover the shared secret
shared_secret_recipient = recipient_key.decapsulate(ciphertext)

print("\n=== Kyber KEM Flow ===")
print(f"Ciphertext             : {ciphertext.hex()}")
print(f"Sender Shared Secret   : {shared_secret_sender.hex()}")
print(f"Recipient Shared Secret: {shared_secret_recipient.hex()}")
assert shared_secret_sender == shared_secret_recipient, "Shared secrets do not match!"


# Generate the signing keypair (private + public)
sign_key = Asym.Cipher.DILITHIUM.keypair.new("dilithium5")

# Sign a message using the private key
message = b"Hello World"
signature = sign_key.sign(message)

# Extract and share only the public key
pub_key_bytes = sign_key.encode_public_key()
# Keys can't be regenerated and try: except: takes more space;
# This can only happen if you do not pass one of the keys when using .decode( ... )
if pub_key_bytes is None:
    raise ValueError("sign_key has no public key")

# Create a new key object with only the public key for verification
verify_key = Asym.Cipher.DILITHIUM.keypair.decode(
    "dilithium5", public_key=pub_key_bytes
)
# Verify the signature using the public key
is_valid = verify_key.sign_verify(message, signature)

print("\n=== Dilithium Signature Flow ===")
print(f"Signature     : {signature.hex()}")
print(f"Signature Valid? {is_valid}")
assert is_valid, "Signature verification failed!"

set_backend(
    [
        Backend.cryptography,
    ]
)

# AES test
crypt_key = Sym.Cipher.AES.key.new(128)
encrypted: bytes = crypt_key.encrypt(b"Hello World", Sym.Padding.PKCS7, Sym.Operation.GCM)
print(encrypted)
decrypted: bytes = crypt_key.decrypt(encrypted, Sym.Padding.PKCS7, Sym.Operation.GCM)
print(decrypted)



"""Key Derivation Functions (KDFs)"""

from .._definitions import (
    PBKDF2HMAC,
    Scrypt,
    HKDF,
    X963,
    ConcatKDF,
    PBKDF1,
    KMAC128,
    KMAC256,
    ARGON2,
    KKDF,
    BCRYPT,
)

__all__ = [
    "PBKDF2HMAC",
    "Scrypt",
    "HKDF",
    "X963",
    "ConcatKDF",
    "PBKDF1",
    "KMAC128",
    "KMAC256",
    "ARGON2",
    "KKDF",
    "BCRYPT",
]



"""TBA"""
from abc import ABCMeta, abstractmethod
import enum

# Internal imports
from .exceptions import NotSupportedError

# Standard typing imports for aps
import typing_extensions as _te
import collections.abc as _a
import typing as _ty

if _ty.TYPE_CHECKING:
    import _typeshed as _tsh
    from .. import GenericLabeledEnum
    from .algos._sym import (KeyEncoding as SymKeyEncoding, Padding as SymPadding,
                             Operation as SymOperation, MessageAuthenticationCode)
    from .algos._asym import (KeyEncoding as ASymKeyEncoding, KeyFormat as ASymKeyFormat,
                              Padding as ASymPadding)
    from .algos import _hash_algorithm
import types as _ts

__all__ = ["Backend", "Cipher", "SymCipher", "AsymCipher", "SingleSerializable", ""]


class Backend(enum.Enum):
    """One of the cryptography backends"""
    cryptography = ("acecurity.crypto._crypto", "Cryptography backend")
    pycryptodomex = ("acecurity.crypto._pycryptox", "PyCryptodomeX backend")
    quantcrypt = ("acecurity.crypto._quantcrypt", "Post-quantum via QuantCrypt")
    argon2_cffi = ("acecurity.crypto._argon2_cffi", "Argon2 via argon2_cffi")
    bcrypt = ("acecurity.crypto._bcrypt", "Bcrypt backend")
    # pynacl = ("acecurity.crypto._pynacl", "NaCl/libsodium backend")  # Unsupported
    # tink = ("acecurity.crypto._tink", "Google Tink backend")  # Unsupported
    # rust_crypto = ("acecurity.crypto._rust", "Rust-backed cryptographic backend")  # Unsupported
    # openssl = ("acecurity.crypto._openssl", "OpenSSL backend")  # Unsupported
    std_lib = ("acecurity.crypto._std_lib", "Standard library fallbacks")


class Cipher(metaclass=ABCMeta):
    """Base class for symmetric and asymmetric ciphers."""

    # cipher: str
    # def decode_key(self, key: bytes, encoding: "SymKeyEncoding" | "ASymKeyEncoding"): ...
    # def __str__(self) -> str:
    #     return self.cipher


class SymCipher(Cipher, metaclass=ABCMeta):
    """Base class for symmetric ciphers."""

    key: _ty.Type["_BaseSymmetricKey"]

    def __str__(self) -> str:
        return self.key.cipher

    # @abstractmethod
    # def new_key(self, __item: _ty.Any) -> "_BaseSymmetricKey":
    #     """Generate or derive a new key from the given input."""
    # @classmethod
    # def decode_key(cls, key: bytes, encoding: "SymKeyEncoding") -> "_BaseSymmetricKey":
    #     """Decode a symmetric key from its encoded representation."""
    #     if cls.key is None:
    #         raise NotSupportedError(f"The {cls()} cipher is not supported by this backend")
    #     if not (isinstance(key, bytes) and isinstance(encoding, "SymKeyEncoding")):
    #         raise ValueError("key needs to be of type bytes and encoding needs to be of type Sym.KeyEncoding")
    #     return cls.key.decode(key, encoding)


class AsymCipher(Cipher, metaclass=ABCMeta):
    """Base class for asymmetric ciphers."""

    keypair: _ty.Type["_BaseAsymmetricKeypair"]

    def __str__(self) -> str:
        return self.keypair.cipher

    # @abstractmethod
    # def new_key(self, __item: _ty.Any) -> "_BaseAsymmetricKeypair":
    #     """Generate or derive a new key from the given input."""
    # @classmethod
    # def decode_private_key(cls, key: bytes, format_: "ASymKeyFormat", encoding: "ASymKeyEncoding") -> "_BaseAsymmetricKeypair":
    #     """Decode an asymmetric key from its encoded representation."""
    #     if cls.keypair is None:
    #         raise NotSupportedError(f"The {cls()} cipher is not supported by this backend")
    #     if not (isinstance(key, bytes) and isinstance(format_, "ASymKeyFormat") and isinstance(encoding, "ASymKeyEncoding")):
    #         raise ValueError("key needs to be of type bytes and encoding needs to be of type ASym.KeyEncoding")
    #     return cls.keypair.decode_private_key(key, format_, encoding)
    # @classmethod
    # def decode_public_key(cls, key: bytes, format_: "ASymKeyFormat", encoding: "ASymKeyEncoding") -> "_BaseAsymmetricKeypair":
    #     """Decode an asymmetric key from its encoded representation."""
    #     if cls.keypair is None:
    #         raise NotSupportedError(f"The {cls()} cipher is not supported by this backend")
    #     if not (isinstance(key, bytes) and isinstance(format_, "ASymKeyFormat") and isinstance(encoding, "ASymKeyEncoding")):
    #         raise ValueError("key needs to be of type bytes and encoding needs to be of type ASym.KeyEncoding")
    #     return cls.keypair.decode_public_key(key, format_, encoding)


class _BaseKey(metaclass=ABCMeta):
    backend: Backend
    cipher: str
    __concrete__: bool = False  # Is flipped true if the set_backend function sets this key, otherwise false

    @abstractmethod
    def new(self, *args: _ty.Any, **kwargs: _ty.Any) -> _te.Self:
        """TBA"""

    @abstractmethod
    def __repr__(self) -> str: ...


class _BaseSymmetricKey(_BaseKey, metaclass=ABCMeta):
    """Abstract base class for basic key types used in symmetric encryption schemes."""


class _BaseAsymmetricKeypair(_BaseKey, metaclass=ABCMeta):
    """Abstract base class for asymmetric key pair operations."""


# Serialization
class SingleSerializable(metaclass=ABCMeta):
    """Interface for encoding and decoding single keys."""

    @classmethod
    @abstractmethod
    def decode(cls, *args: _ty.Any, **kwargs: _ty.Any) -> _te.Self:
        """TBA"""

    @abstractmethod
    def encode(self, *args: _ty.Any, **kwargs: _ty.Any) -> _ty.Any:
        """TBA"""


class SymKeySerializable(SingleSerializable, metaclass=ABCMeta):
    """Interface for encoding and decoding symmetric keys."""

    @classmethod
    @abstractmethod
    def decode(cls, key: bytes, encoding: "SymKeyEncoding") -> _te.Self:
        """
        Deserialize a key from its encoded byte representation.
        :param key: The encoded key bytes.
        :param encoding: The encoding format used (e.g., RAW, BASE64).
        :return: An instance of the key class constructed from the input.
        """

    @abstractmethod
    def encode(self, encoding: "SymKeyEncoding") -> bytes:
        """
        Serialize the key into a specific encoding format.
        :param encoding: The format to encode the key into.
        :return: The encoded key.
        """


class DoubleSerializable(metaclass=ABCMeta):
    """Interface for encoding and decoding double keys."""

    @classmethod
    @abstractmethod
    def decode_private_key(cls, *args: _ty.Any, **kwargs: _ty.Any) -> _te.Self:
        """
        Deserialize a private key from the given encoded data.
        :param data:: Encoded private key data.
        :param format_: The key format (e.g., PKCS8).
        :param encoding: The key encoding (e.g., PEM, DER).
        :param password: Password used to decrypt the private key if encrypted.
        :return: A new instance of the key pair object containing the private key.
        """

    @classmethod
    @abstractmethod
    def decode_public_key(cls, *args: _ty.Any, **kwargs: _ty.Any) -> _te.Self:
        """
        Deserialize a public key from the given encoded data.
        :param data: Encoded public key data.
        :param format_: The key format (e.g., SubjectPublicKeyInfo).
        :param encoding: The key encoding (e.g., PEM, DER).
        :return: An instance containing only the public key.
        """

    @abstractmethod
    def encode_private_key(self, *args: _ty.Any, **kwargs: _ty.Any) -> _ty.Any:
        """
        Serialize the private key to the given format and encoding.
        :param format_: Format to use for the private key.
        :param encoding: Encoding (PEM, DER, etc.).
        :param password: If provided, encrypts the private key with this password.
        :return: The encoded private key.
        """

    @abstractmethod
    def encode_public_key(self, *args: _ty.Any, **kwargs: _ty.Any) -> _ty.Any:
        """
        Serialize the public key to the given format and encoding.
        :param format_: Format to use for the public key.
        :param encoding: Encoding (PEM, DER, etc.).
        :return: The encoded public key.
        """


class AsymKeySerializable(DoubleSerializable, metaclass=ABCMeta):
    """Interface for encoding and decoding asymmetric key pairs."""

    @classmethod
    @abstractmethod
    def decode_private_key(
        cls,
        data: bytes,
        format_: "ASymKeyFormat",
        encoding: "ASymKeyEncoding",
        password: bytes | None = None,
    ) -> _te.Self:
        """
        Deserialize a private key from the given encoded data.
        :param data:: Encoded private key data.
        :param format_: The key format (e.g., PKCS8).
        :param encoding: The key encoding (e.g., PEM, DER).
        :param password: Password used to decrypt the private key if encrypted.
        :return: A new instance of the key pair object containing the private key.
        """

    @classmethod
    @abstractmethod
    def decode_public_key(
        cls, data: bytes, format_: "ASymKeyFormat", encoding: "ASymKeyEncoding"
    ) -> _te.Self:
        """
        Deserialize a public key from the given encoded data.
        :param data: Encoded public key data.
        :param format_: The key format (e.g., SubjectPublicKeyInfo).
        :param encoding: The key encoding (e.g., PEM, DER).
        :return: An instance containing only the public key.
        """

    @abstractmethod
    def encode_private_key(
        self,
        format_: "ASymKeyFormat",
        encoding: "ASymKeyEncoding",
        password: bytes | None = None,
    ) -> bytes:
        """
        Serialize the private key to the given format and encoding.
        :param format_: Format to use for the private key.
        :param encoding: Encoding (PEM, DER, etc.).
        :param password: If provided, encrypts the private key with this password.
        :return: The encoded private key.
        """

    @abstractmethod
    def encode_public_key(
        self, format_: "ASymKeyFormat", encoding: "ASymKeyEncoding"
    ) -> bytes:
        """
        Serialize the public key to the given format and encoding.
        :param format_: Format to use for the public key.
        :param encoding: Encoding (PEM, DER, etc.).
        :return: The encoded public key.
        """


# Capabilities
class SymKeyEncryptable(metaclass=ABCMeta):
    """Interface for symmetric encryption and decryption."""

    @abstractmethod
    def encrypt(
        self,
        plain_bytes: bytes,
        padding: "SymPadding",
        mode: "SymOperation",
        /,
        *,
        auto_pack: bool = True,
    ) -> bytes | dict[str, bytes]:
        """
        Encrypt a block of plaintext bytes.
        :param plain_bytes: The plaintext data to encrypt.
        :param padding: The padding scheme to apply.
        :param mode: The cipher operation mode (e.g., CBC, GCM).
        :param auto_pack: If all data should be packed into one bytes object.
        :return: The encrypted output, either as raw bytes or a structured dictionary depending on `auto_pack`.
        """

    @abstractmethod
    def decrypt(
        self,
        cipher_bytes_or_dict: bytes | dict[str, bytes],
        padding: "SymPadding",
        mode: "SymOperation",
        /,
        *,
        auto_pack: bool = True,
    ) -> bytes:
        """
        Decrypt ciphertext bytes into plaintext.
        :param cipher_bytes_or_dict: The ciphertext input, either as raw bytes or a dictionary containing metadata.
        :param padding: The padding scheme used during encryption.
        :param mode: The cipher operation mode.
        :param auto_pack: If True, expects bytes. Defaults to True.
        :return: The decrypted plaintext bytes.
        """


class SymKeyMACCapable(metaclass=ABCMeta):
    """Interface for MAC generation and verification."""

    @abstractmethod
    def generate_mac(self, data: bytes, auth_type: "MessageAuthenticationCode") -> bytes:
        """
        Generate a message authentication code.

        :param data: Message to authenticate.
        :param auth_type: Type of MAC (e.g., HMAC, CMAC).
        :return: Generated MAC bytes.
        """

    @abstractmethod
    def verify_mac(
        self, data: bytes, mac: bytes, auth_type: "MessageAuthenticationCode"
    ) -> bool:
        """
        Verify a message authentication code.

        :param data: Original message.
        :param mac: Expected MAC.
        :param auth_type: MAC type used.
        :return: True if MAC is valid, else False.
        """


class AsymKeyEncryptable(metaclass=ABCMeta):
    """Interface for public-key encryption and decryption."""

    @abstractmethod
    def encrypt(self, plain_bytes: bytes, padding: "ASymPadding") -> bytes:
        """
        Encrypt data using the public key and specified padding scheme.
        :param plain_bytes: Data to encrypt.
        :param padding: Padding scheme to use.
        :return: Encrypted ciphertext.
        """

    @abstractmethod
    def decrypt(self, cipher_bytes: bytes, padding: "ASymPadding") -> bytes:
        """
        Decrypt data using the private key and specified padding scheme.
        :param cipher_bytes: Data to decrypt.
        :param padding: Padding scheme used during encryption.
        :return: Decrypted plaintext.
        """


class Signable(metaclass=ABCMeta):
    """Interface for signing and verifying data with asymmetric keys."""

    @abstractmethod
    def sign(self, *args: _ty.Any, **kwargs: _ty.Any) -> bytes:
        """TBA"""

    @abstractmethod
    def sign_verify(self, *args: _ty.Any, **kwargs: _ty.Any) -> bool:
        """TBA"""


class AsymKeySignable(Signable, metaclass=ABCMeta):
    """Interface for signing and verifying data with asymmetric keys."""

    @abstractmethod
    def sign(self, data: bytes, padding: "ASymPadding") -> bytes:
        """
        Generate a digital signature for the given data using the private key.
        :param data: Data to sign.
        :param padding: Padding scheme for signing.
        :return: Signature.
        """

    @abstractmethod
    def sign_verify(self, data: bytes, signature: bytes, padding: "ASymPadding") -> bool:
        """
        Verify a digital signature using the public key.
        :param data: Original data that was signed.
        :param signature: The signature to verify.
        :param padding: Padding scheme used during signing.
        :return: True if the signature is valid, False otherwise.
        """


class AsymKeyKeyExchangeable(metaclass=ABCMeta):
    """Interface for performing a key exchange operation."""

    @abstractmethod
    def key_exchange(self, peer_public_key: _te.Self) -> bytes:
        """
        Perform a key exchange operation using this private key and a peer's public key.
        :param peer_public_key: The peer's public key.
        :return: The shared secret resulting from the key exchange.
        """


class AsymEncapsulatable(metaclass=ABCMeta):
    """Interface for key encapsulation mechanisms (KEM)."""

    @abstractmethod
    def encapsulate(self) -> tuple[bytes, bytes]:
        """
        Generate a key encapsulation.
        :return: (ciphertext, shared_secret)
        """

    @abstractmethod
    def decapsulate(self, ciphertext: bytes) -> bytes:
        """
        Derive shared secret from encapsulated ciphertext.
        :param ciphertext: The ciphertext from peer.
        :return: The shared secret.
        """


# Advanced
class Fingerprintable(metaclass=ABCMeta):
    """Allows generation of fingerprints from public keys."""

    @abstractmethod
    def fingerprint(self, hash_alg: str = "sha256") -> str:
        """
        Return a hash-based identifier of the key.

        :param hash_alg: Hash function to use.
        :return: Hex-encoded fingerprint.
        """


class Zeroizable(metaclass=ABCMeta):
    """Interface for zeroizing sensitive key material."""

    @abstractmethod
    def zeroize(self) -> None:
        """Securely wipes key material from memory."""


class _BASIC_HASHER(metaclass=ABCMeta):
    """Abstract base class for hashing algorithms."""

    backend: Backend

    @abstractmethod
    def hash(self, *args: _ty.Any, **kwargs: _ty.Any) -> bytes:
        """
        Generate a hash of the given input data.
        :param to_hash: The input data to hash.
        :param text_ids: Add text ids to the resulting hash
        :return: The resulting hash digest.
        """

    @abstractmethod  # , text_ids: bool = True
    def verify(self, to_verify: bytes, original_hash: bytes, *args: _ty.Any, **kwargs: _ty.Any) -> bool:
        """
        Verify that the hash of the given input matches the provided hash.
        :param to_verify: The input data to hash and verify.
        :param original_hash: The original hash to compare against.
        :return: True if the hash matches, False otherwise.
        """


_HASHLITERAL = _ty.Literal[
    4,
    5,
    6,
    7,
    8,
    9,
    10,
    11,
    12,
    13,
    14,
    15,
    16,
    17,
    18,
    19,
    20,
    21,
    22,
    23,
    24,
    25,
    26,
    27,
    28,
    29,
    30,
    31,
    32,
    33,
]
_HASHALGOLITERAL = _ty.Literal[
    "sha1",
    "sha224",
    "sha256",
    "sha384",
    "sha512",
    "sha512_224",
    "sha512_256",
    "sha3_224",
    "sha3_256",
    "sha3_384",
    "sha3_512",
    "sha3_shake_128",
    "sha3_shake_256",
    "sha3_turbo_shake_128",
    "sha3_turbo_shake_256",
    "sha3_kangaroo_twelve",
    "sha3_tuple_hash_128",
    "sha3_tuple_hash_256",
    "sha3_keccak",
    "sha3_cshake_128",
    "sha3_cshake_256",
    "blake2b",
    "blake2s",
    "md2",
    "md4",
    "md5",
    "sm3",
    "ripemd160",
    "bcrypt",
    "argon2",
]
_HASHID_TO_STRING: dict[_HASHLITERAL, str] = {
    4: "sha1",
    5: "sha224",
    6: "sha256",
    7: "sha384",
    8: "sha512",
    9: "sha512_224",
    10: "sha512_256",
    11: "sha3_224",
    12: "sha3_256",
    13: "sha3_384",
    14: "sha3_512",
    15: "sha3_shake_128",
    16: "sha3_shake_256",
    17: "sha3_turbo_shake_128",
    18: "sha3_turbo_shake_256",
    19: "sha3_kangaroo_twelve",
    20: "sha3_tuple_hash_128",
    21: "sha3_tuple_hash_256",
    22: "sha3_keccak",
    23: "blake2b",
    24: "blake2s",
    25: "md2",
    26: "md4",
    27: "md5",
    28: "sm3",
    29: "ripemd160",
    30: "bcrypt",
    31: "argon2",
    32: "cshake_128",
    33: "cshake_256",
}


class _HASHER_BACKEND(_BASIC_HASHER):
    _MAPPING: dict[
        str, tuple[_a.Callable[[bytes], bytes], _a.Callable[[bytes, bytes], bool]]
    ] = {}

    def __init__(self, hash_algorithm: _HASHALGOLITERAL):  # , hash_type: _HASHLITERAL
        # self.hash_type: _HASHLITERAL = hash_type
        # algorithm: str  #  | None = _HASHID_TO_STRING.get(hash_type)
        if hash_algorithm not in list(_HASHID_TO_STRING.values()):  # not algorithm:
            raise ValueError(f"Unsupported hash algorithm {hash_algorithm}")
        self.algorithm: str = hash_algorithm

    @classmethod
    def verify_unknown(
        cls,
        to_verify: bytes,
        original_hash: bytes,
        /,
        fallback_algorithm: str = "sha256",
        text_ids: bool = True,
    ) -> bool:
        algorithm: str = fallback_algorithm
        if text_ids:
            _, algo, original_hash = original_hash.split(b"$", maxsplit=2)
            algorithm = algo.decode("utf-8")
        impl: (
            tuple[_a.Callable[[bytes], bytes], _a.Callable[[bytes, bytes], bool]] | None
        ) = cls._MAPPING.get(algorithm)
        if impl is None:
            raise NotSupportedError(
                f"Could not verify unknown algorithm '{algorithm}'."
            )
        return impl[1](to_verify, original_hash)

    def hash(self, to_hash: bytes, /, text_ids: bool = True) -> bytes:
        impl: (
            tuple[_a.Callable[[bytes], bytes], _a.Callable[[bytes, bytes], bool]] | None
        ) = self._MAPPING.get(self.algorithm)
        if impl is None:
            raise NotSupportedError(
                f"The {self} hash is not supported by this backend"
            )
        result: bytes = impl[0](to_hash)
        if text_ids:
            result = b"$" + self.algorithm.encode("utf-8") + b"$" + result
        return result

    def verify(self, to_verify: bytes, original_hash: bytes, *, text_ids: bool = True) -> bool:
        impl: (
            tuple[_a.Callable[[bytes], bytes], _a.Callable[[bytes, bytes], bool]] | None
        ) = self._MAPPING.get(self.algorithm)
        if impl is None:
            raise NotSupportedError(
                f"The {self} hash is not supported by this backend"
            )
        if text_ids:
            _, _, original_hash = original_hash.split(b"$", maxsplit=2)
        return impl[1](to_verify, original_hash)

    def __str__(self) -> str:
        return self.algorithm

    def __repr__(self) -> str:
        return f"<HashBackend alg={self.algorithm}>"


class _HASHER_WITH_LEN_BACKEND(_HASHER_BACKEND):
    def hash(self, to_hash: bytes, hash_len: int, *, text_ids: bool = True) -> bytes:
        impl: (
            tuple[_a.Callable[[bytes, int], bytes], _a.Callable[[bytes, bytes], bool]]
            | None
        ) = _HASHER_BACKEND._MAPPING.get(self.algorithm)
        if impl is None:
            raise NotSupportedError(
                f"The {self} hash is not supported by this backend"
            )
        result: bytes = impl[0](to_hash, hash_len)
        if text_ids:
            result = b"$" + self.algorithm.encode("utf-8") + b"$" + result
        return result


# TODO: REMOVE WHEN THE CIRCULAR IMPORTS ARE FIXED
class _hash_algorithm:
    class SHA2:
        SHA256 = _HASHER_BACKEND("sha256")

    MD5 = _HASHER_BACKEND("md5")


class _BASIC_KEY_DERIVATION_FUNC(metaclass=ABCMeta):
    """Abstract base class for key derivation functions (KDFs)."""

    backend: Backend
    _IMPL: _a.Callable[..., bytes] | None = None

    # @classmethod
    # @abstractmethod
    # def derive(cls, password: bytes) -> bytes:
    #     """
    #     Derive a cryptographic key from a password.
    #     :param password: The input password or secret to derive the key from.
    #     :return: The derived key of the requested length.
    #     """


kdfs = [
    "pbkdf2hmac",
    "scrypt",
    "hkdf",
    "x963",
    "concatkdf",
    "pbkdf1",
    "kmac128",
    "kmac256",
    "argon2",
    "kkdf",
    "bcrypt",
]


class PBKDF2HMAC(_BASIC_KEY_DERIVATION_FUNC):
    """Password-Based Key Derivation Function 2"""

    _IMPL: _a.Callable[[bytes, bytes, int, int, str], bytes] | None = None

    @classmethod
    def derive(cls, password: bytes, *, salt: bytes, length: int = 32, iterations: int = 100_000,
               hash_alg: _HASHER_BACKEND = _hash_algorithm.SHA2.SHA256) -> bytes:
        """

        :param password:
        :param salt:
        :param length:
        :param iterations:
        :param hash_alg:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, salt, length, iterations, hash_alg.algorithm)

    def __str__(self) -> str:
        return "PBKDF2HMAC"


class Scrypt(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: _a.Callable[[bytes, bytes | None, int, int, int, int], bytes] | None = None

    @classmethod
    def derive(
        cls,
        password: bytes,
        *,
        salt: bytes,
        length: int = 32,
        n: int = 2**14,
        r: int = 8,
        p: int = 1,
    ) -> bytes:
        """

        :param password:
        :param salt:
        :param length:
        :param n:
        :param r:
        :param p:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, salt, length, n, r, p)

    def __str__(self) -> str:
        return "Scrypt"


class HKDF(_BASIC_KEY_DERIVATION_FUNC):
    """HMAC-based Extract-and-Expand Key Derivation Function"""

    _IMPL: _a.Callable[[bytes, bytes | None, bytes, int, str], bytes] | None = None

    @classmethod
    def derive(
        cls,
        password: bytes,
        *,
        salt: bytes,
        info: bytes = b"",
        length: int = 32,
        hash_alg: _HASHER_BACKEND = _hash_algorithm.SHA2.SHA256,
    ) -> bytes:
        """

        :param password:
        :param salt:
        :param info:
        :param length:
        :param hash_alg:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, salt, info, length, hash_alg.algorithm)

    def __str__(self) -> str:
        return "HKDF"


class X963(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: _a.Callable[[bytes, int, bytes, str], bytes] | None = None

    @classmethod
    def derive(
        cls,
        password: bytes,
        *,
        length: int = 32,
        otherinfo: bytes,
        hash_alg: _HASHER_BACKEND = _hash_algorithm.SHA2.SHA256,
    ) -> bytes:
        """

        :param password:
        :param length:
        :param otherinfo:
        :param hash_alg:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, length, otherinfo, hash_alg.algorithm)

    def __str__(self) -> str:
        return "X963"


class ConcatKDF(X963):
    """TBA"""

    def __str__(self) -> str:
        return "ConcatKDF"


class PBKDF1(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: _a.Callable[[bytes, bytes, int, int, str], bytes] | None = None

    @classmethod
    def derive(
        cls,
        password: bytes,
        *,
        salt: bytes,
        length: int = 32,
        iterations: int = 1_000,
        hash_alg: _HASHER_BACKEND = _hash_algorithm.MD5,
    ) -> bytes:
        """

        :param password:
        :param salt:
        :param length:
        :param iterations:
        :param hash_alg:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, salt, length, iterations, hash_alg.algorithm)

    def __str__(self) -> str:
        return "PBKDF1"


class KMAC128(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: _a.Callable[[bytes, int, bytes, bytes], bytes] | None = None

    @classmethod
    def derive(
        cls,
        password: bytes,
        *,
        length: int,
        key: bytes = b"",
        customization: bytes = b"",
    ) -> bytes:
        """

        :param password:
        :param length:
        :param key:
        :param customization:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, length, key, customization)

    def __str__(self) -> str:
        return "KMAC128"


class KMAC256(KMAC128):
    def __str__(self) -> str:
        return "KMAC256"


class ARGON2(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: (
        _a.Callable[
            [bytes, bytes, int, int, int, int, _ty.Literal["i", "d", "id"]], bytes
        ]
        | None
    ) = None

    @classmethod
    def derive(
        cls,
        password: bytes,
        *,
        salt: bytes,
        length: int = 32,
        time_cost: int = 2,
        memory_cost: int = 102_400,
        parallelism: int = 8,
        variant: _ty.Literal["i", "d", "id"] = "i",
    ) -> bytes:
        """

        :param password:
        :param salt:
        :param length:
        :param time_cost:
        :param memory_cost:
        :param parallelism:
        :param variant:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(
            password, salt, length, time_cost, memory_cost, parallelism, variant
        )

    def __str__(self) -> str:
        return "ARGON2"


class KKDF(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: _a.Callable[[bytes, int, bytes], bytes] | None = None

    @classmethod
    def derive(cls, password: bytes, *, length: int = 32, salt: bytes = b"") -> bytes:
        """

        :param password:
        :param length:
        :param salt:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, length, salt)

    def __str__(self) -> str:
        return "KKDF"


class BCRYPT(_BASIC_KEY_DERIVATION_FUNC):
    """TBA"""

    _IMPL: _a.Callable[[bytes, bytes, int, int], bytes] | None = None

    @classmethod
    def derive(
        cls, password: bytes, /, salt: bytes, rounds: int = 12, length: int = 32
    ) -> bytes:
        """

        :param password:
        :param salt:
        :param rounds:
        :param length:
        :return:
        """
        if cls._IMPL is None:
            raise NotSupportedError(
                f"The {cls()} KDF is not supported by this backend"
            )  # Throw tantrum
        return cls._IMPL(password, salt, rounds, length)

    def __str__(self) -> str:
        return "BCRYPT"


_AES_KEYSIZES = (128, 192, 256)
_AES_KEYLITERAL = _ty.Literal[128, 192, 256]


class _AES_KEYTYPE(
    _BaseSymmetricKey,
    SymKeyEncryptable,
    SymKeyMACCapable,
    SymKeySerializable,
    metaclass=ABCMeta,
):
    cipher = "AES"

    def __init__(
        self, key_size: _AES_KEYLITERAL, pwd: _ty.Optional[bytes | str]
    ) -> None: ...

    @classmethod
    def new(
        cls, pwd_or_keysize: _AES_KEYLITERAL | tuple[_AES_KEYLITERAL, str | bytes]
    ) -> _te.Self:
        """TBA"""
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _AES_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _AES_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _AES_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_AES_KEYLITERAL, key_size_or_key)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_AES_KEYLITERAL} | tuple[{_AES_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class AES(SymCipher):
    """Advanced Encryption Standard"""

    key: _ty.Type[_AES_KEYTYPE] = _AES_KEYTYPE


class _ChaCha20_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "ChaCha20"

    def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None: ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(pwd)


class ChaCha20(SymCipher):
    key: _ty.Type[_ChaCha20_KEYTYPE] = _ChaCha20_KEYTYPE


class _TripleDES_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "TripleDES"

    def __init__(
        self, key_size: _ty.Literal[192], pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(192, pwd)


class TripleDES(SymCipher):
    key: _ty.Type[_TripleDES_KEYTYPE] = _TripleDES_KEYTYPE


_Blowfish_KEYSIZES = (128, 256)
_Blowfish_KEYLITERAL = _ty.Literal[128, 256]


class _Blowfish_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "Blowfish"

    def __init__(
        self, key_size: _ty.Literal[128, 256], pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls,
        pwd_or_keysize: _Blowfish_KEYLITERAL | tuple[_Blowfish_KEYLITERAL, str | bytes],
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _Blowfish_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _Blowfish_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _Blowfish_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_Blowfish_KEYLITERAL, key_size_or_key)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_Blowfish_KEYLITERAL} | tuple[{_Blowfish_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class Blowfish(SymCipher):
    key: _ty.Type[_Blowfish_KEYTYPE] = _Blowfish_KEYTYPE


_CAST5_KEYSIZES = (40, 128)
_CAST5_KEYLITERAL = _ty.Literal[40, 128]


class _CAST5_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "CAST5"

    def __init__(
        self, key_size: _ty.Literal[40, 128], pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls, pwd_or_keysize: _CAST5_KEYLITERAL | tuple[_CAST5_KEYLITERAL, str | bytes]
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _CAST5_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _CAST5_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _CAST5_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), str)
        ):
            pass  # key_size = _ty.cast(_CAST5_KEYLITERAL, __item)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_CAST5_KEYLITERAL} | tuple[{_CAST5_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class CAST5(SymCipher):
    key: _ty.Type[_CAST5_KEYTYPE] = _CAST5_KEYTYPE


class _ARC4_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "ARC4"

    def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None: ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(pwd)


class ARC4(SymCipher):
    key: _ty.Type[_ARC4_KEYTYPE] = _ARC4_KEYTYPE


_Camellia_KEYSIZES = (128, 192, 256)
_Camellia_KEYLITERAL = _ty.Literal[128, 192, 256]


class _Camellia_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "Camellia"

    def __init__(
        self, key_size: _Camellia_KEYLITERAL, pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls,
        pwd_or_keysize: _Camellia_KEYLITERAL | tuple[_Camellia_KEYLITERAL, str | bytes],
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _Camellia_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _Camellia_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _Camellia_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_Camellia_KEYLITERAL, pwd_or_keysize)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_Camellia_KEYLITERAL} | tuple[{_Camellia_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class Camellia(SymCipher):
    key: _ty.Type[_Camellia_KEYTYPE] = _Camellia_KEYTYPE


class _IDEA_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "IDEA"

    def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None: ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(pwd)


class IDEA(SymCipher):
    key: _ty.Type[_IDEA_KEYTYPE] = _IDEA_KEYTYPE


class _SEED_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "SEED"

    def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None: ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(pwd)


class SEED(SymCipher):
    key: _ty.Type[_SEED_KEYTYPE] = _SEED_KEYTYPE


class _SM4_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "SM4"

    def __init__(self, pwd: _ty.Optional[bytes | str] = None): ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(pwd)


class SM4(SymCipher):
    key: _ty.Type[_SM4_KEYTYPE] = _SM4_KEYTYPE


class _DES_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "DES"

    def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None: ...

    @classmethod
    def new(cls, pwd: bytes | str | None = None) -> _te.Self:
        """
        if isinstance(pwd, (bytes, str)):
            if len(pwd) * 8 != 64:
                raise ValueError(f"Can't accept key of size '{len(pwd) * 8}' for the {cls()} cipher, please use a 64-bit long one")
            elif pwd is not None:
                raise ValueError(f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'")
        """
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if not (isinstance(pwd, (bytes, str)) or pwd is None):
            raise ValueError(
                f"pwd needs to be of type 'bytes | str | None', and not '{repr(pwd)}'"
            )
        return cls(pwd)


class DES(SymCipher):
    key: _ty.Type[_DES_KEYTYPE] = _DES_KEYTYPE


_ARC2_KEYSIZES = (40, 64, 128)
_ARC2_KEYLITERAL = _ty.Literal[40, 64, 128]


class _ARC2_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "ARC2"

    def __init__(
        self, key_size: _ARC2_KEYLITERAL = 128, pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls, pwd_or_keysize: _ARC2_KEYLITERAL | tuple[_ARC2_KEYLITERAL, str | bytes]
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _ARC2_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _ARC2_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _ARC2_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_ARC2_KEYLITERAL, pwd_or_keysize)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_ARC2_KEYLITERAL} | tuple[{_ARC2_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )

        return cls(key_size, pwd)


class ARC2(SymCipher):
    key: _ty.Type[_ARC2_KEYTYPE] = _ARC2_KEYTYPE


_Salsa20_KEYSIZES = (128, 256)
_Salsa20_KEYLITERAL = _ty.Literal[128, 256]


class _Salsa20_KEYTYPE(
    _BaseSymmetricKey, SymKeyEncryptable, SymKeySerializable, metaclass=ABCMeta
):
    cipher = "Salsa20"

    def __init__(
        self, key_size: _Salsa20_KEYLITERAL = 256, pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls,
        pwd_or_keysize: _Salsa20_KEYLITERAL | tuple[_Salsa20_KEYLITERAL, str | bytes],
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _Salsa20_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _Salsa20_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _Salsa20_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_Salsa20_KEYLITERAL, pwd_or_keysize)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_Salsa20_KEYLITERAL} | tuple[{_Salsa20_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class Salsa20(SymCipher):
    key: _ty.Type[_Salsa20_KEYTYPE] = _Salsa20_KEYTYPE


_RSA_KEYSIZES = (512, 768, 1024, 2048, 3072, 4096, 8192, 15360)
_RSA_KEYLITERAL = _ty.Literal[512, 768, 1024, 2048, 3072, 4096, 8192, 15360]


class _RSA_KEYPAIRTYPE(
    _BaseAsymmetricKeypair,
    AsymKeyEncryptable,
    AsymKeySignable,
    AsymKeySerializable,
    metaclass=ABCMeta,
):
    cipher = "RSA"

    def __init__(
        self, key_size: _RSA_KEYLITERAL, pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls, pwd_or_keysize: _RSA_KEYLITERAL | tuple[_RSA_KEYLITERAL, str | bytes]
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _RSA_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _RSA_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _RSA_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_RSA_KEYLITERAL, pwd_or_keysize)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_RSA_KEYLITERAL} | tuple[{_RSA_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class RSA(AsymCipher):
    """Rivest-Shamir-Adleman"""

    keypair: _ty.Type[_RSA_KEYPAIRTYPE] = _RSA_KEYPAIRTYPE


_DSA_KEYSIZES = (1024, 2048, 3072)
_DSA_KEYLITERAL = _ty.Literal[1024, 2048, 3072]


class _DSA_KEYPAIRTYPE(
    _BaseAsymmetricKeypair, AsymKeySignable, AsymKeySerializable, metaclass=ABCMeta
):
    cipher = "DSA"

    def __init__(
        self, key_size: _DSA_KEYLITERAL, pwd: _ty.Optional[bytes | str] = None
    ) -> None: ...

    @classmethod
    def new(
        cls, pwd_or_keysize: _DSA_KEYLITERAL | tuple[_DSA_KEYLITERAL, str | bytes]
    ) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        pwd: str | bytes | None = None
        key_size: _DSA_KEYLITERAL

        if isinstance(pwd_or_keysize, int) and pwd_or_keysize in _DSA_KEYSIZES:
            key_size = pwd_or_keysize
        elif (
            isinstance(pwd_or_keysize, tuple)
            and len(pwd_or_keysize) == 2
            and isinstance((key_size := pwd_or_keysize[0]), int)
            and key_size in _RSA_KEYSIZES
            and isinstance((pwd := pwd_or_keysize[1]), (str, bytes))
        ):
            pass  # key_size = _ty.cast(_DSA_KEYLITERAL, pwd_or_keysize)
        else:
            raise ValueError(
                f"pwd_or_keysize needs to be of type '{_DSA_KEYLITERAL} | tuple[{_DSA_KEYLITERAL}, str | bytes]', and not '{repr(pwd_or_keysize)}'"
            )
        return cls(key_size, pwd)


class DSA(AsymCipher):
    """Digital Signature Algorithm"""

    keypair: _ty.Type[_DSA_KEYPAIRTYPE] = _DSA_KEYPAIRTYPE


class ECCCurve(enum.Enum):
    """Elliptic key functions"""

    SECP192R1 = (None, "")
    SECP224R1 = (None, "")
    SECP256K1 = (None, "")
    SECP256R1 = (None, "")  # Default
    SECP384R1 = (None, "")
    SECP521R1 = (None, "")
    SECT163K1 = (None, "")
    SECT163R2 = (None, "")
    SECT233K1 = (None, "")
    SECT233R1 = (None, "")
    SECT283K1 = (None, "")
    SECT283R1 = (None, "")
    SECT409K1 = (None, "")
    SECT409R1 = (None, "")
    SECT571K1 = (None, "")
    SECT571R1 = (None, "")


class ECCType(enum.Enum):
    """How the signing is done, heavily affects the performance, key generation and what you can do with it"""

    ECDSA = (None, "Elliptic Curve Digital Signature Algorithm")
    Ed25519 = (None, "")
    Ed448 = (None, "")
    X25519 = (None, "")
    X448 = (None, "")


class _ECC_KEYPAIRTYPE(
    _BaseAsymmetricKeypair,
    AsymKeySignable,
    AsymKeyKeyExchangeable,
    AsymKeySerializable,
    metaclass=ABCMeta,
):
    cipher = "ECC"

    def __init__(
        self,
        ecc_type: ECCType = ECCType.ECDSA,
        ecc_curve: ECCCurve | None = ECCCurve.SECP256R1,
        pwd: _ty.Optional[bytes | str] = None,
    ) -> None: ...

    @classmethod
    def new(cls, __item: _ty.Any) -> _te.Self:
        """
        Creates a new ECC key depending on the provided input.

        Accepted forms:
        - ECC.new((ECCType.ECDSA, ECCCurve.SECP384R1))      # For ECDSA
        - ECC.new((ECCType.X25519, None))                   # For optimized curves
        - ECC.new((ECCType.ECDSA, ECCCurve.SECP384R1, b"..."))  # With private key
        """
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )

        # If it's a tuple, unpack it
        if isinstance(__item, tuple):
            if len(__item) == 2:
                ecc_type, arg = __item
                if ecc_type == ECCType.ECDSA:
                    return cls.ecdsa_key(ecc_curve=arg)
                else:
                    return cls.optimized_key(ecc_type=ecc_type)
            elif len(__item) == 3:
                ecc_type, curve, priv = __item
                if ecc_type == ECCType.ECDSA:
                    return cls.ecdsa_key(ecc_curve=curve, private_key=priv)
                else:
                    return cls.optimized_key(ecc_type=ecc_type, private_key=priv)

        raise ValueError(
            f"{cls.__name__}.new() requires a tuple of the form:\n"
            f"  (ECCType.ECDSA, ECCCurve)\n"
            f"  (ECCType.OTHER, None)\n"
            f"  or with an optional private key as third item."
        )

    @classmethod
    def ecdsa_key(
        cls,
        ecc_curve: ECCCurve = ECCCurve.SECP256R1,
        private_key: bytes | str | None = None,
    ) -> _te.Self:
        """Elliptic Curve Digital Signature Algorithm"""
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        elif not isinstance(ecc_curve, ECCCurve):
            raise ValueError(
                f"ecc_curve needs to be of type 'ECCCurve', and not '{repr(ecc_curve)}'"
            )
        elif not (isinstance(private_key, (str, bytes)) or private_key is None):
            raise ValueError(
                f"private_key needs to be of type 'bytes | str | None', and not '{repr(private_key)}'"
            )
        return cls(ECCType.ECDSA, ecc_curve, private_key)

    @classmethod
    def optimized_key(
        cls, ecc_type: ECCType, private_key: bytes | str | None = None
    ) -> _te.Self:
        """
        TBA
        :param ecc_type:
        :param private_key:
        :return:
        """
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        elif not isinstance(ecc_type, ECCType):
            raise ValueError(
                f"ecc_type needs to be of type 'ECCType', and not '{repr(ecc_type)}'"
            )
        elif not (isinstance(private_key, (str, bytes)) or private_key is None):
            raise ValueError(
                f"private_key needs to be of type 'bytes | str | None', and not '{repr(private_key)}'"
            )
        elif ecc_type == ECCType.ECDSA:
            raise ValueError("Please use ECC.ecdsa_key to generate ECDSA keys")
        return cls(ecc_type, None, private_key)


class ECC(AsymCipher):
    """Elliptic Curve Cryptography"""

    keypair: _ty.Type[_ECC_KEYPAIRTYPE] = _ECC_KEYPAIRTYPE
    Curve = ECCCurve
    Type = ECCType


_KYBER_MODES = ("kyber512", "kyber768", "kyber1024")
_KYBER_MODELITERAL = _ty.Literal["kyber512", "kyber768", "kyber1024"]


class _KYBER_KEYPAIRTYPE(
    _BaseAsymmetricKeypair,
    AsymEncapsulatable,
    # SingleSerializable,
    metaclass=ABCMeta,
):
    cipher = "KYBER"

    def __init__(self, mode: _KYBER_MODELITERAL) -> None: ...
    @classmethod
    def new(cls, mode: _KYBER_MODELITERAL) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if mode not in _KYBER_MODES:
            raise ValueError(f"mode must be one of {_KYBER_MODES}, not '{mode}'")
        return cls(mode)

    @classmethod
    @abstractmethod
    def decode(
        cls,
        mode: _KYBER_MODELITERAL,
        *,
        public_key: bytes | None = None,
        private_key: bytes | None = None,
    ) -> _te.Self: ...
    @abstractmethod
    def encode_private_key(self) -> bytes | None: ...
    @abstractmethod
    def encode_public_key(self) -> bytes | None: ...


class KYBER(AsymCipher):
    keypair: _ty.Type[_KYBER_KEYPAIRTYPE] = _KYBER_KEYPAIRTYPE


_DILITHIUM_MODES = ("dilithium2", "dilithium3", "dilithium5")
_DILITHIUM_MODELITERAL = _ty.Literal["dilithium2", "dilithium3", "dilithium5"]


class _DILITHIUM_KEYPAIRTYPE(
    _BaseAsymmetricKeypair,
    Signable,
    # SingleSerializable,
    metaclass=ABCMeta,
):
    cipher = "DILITHIUM"

    def __init__(self, mode: _DILITHIUM_MODELITERAL) -> None: ...
    @classmethod
    def new(cls, mode: _DILITHIUM_MODELITERAL) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(
                f"The {cls.cipher} cipher is not supported by this backend"
            )
        if mode not in _DILITHIUM_MODES:
            raise ValueError(f"mode must be one of {_DILITHIUM_MODES}, not '{mode}'")
        return cls(mode)

    @classmethod
    @abstractmethod
    def decode(
        cls,
        mode: _DILITHIUM_MODELITERAL,
        *,
        public_key: bytes | None = None,
        private_key: bytes | None = None,
    ) -> _te.Self: ...
    @abstractmethod
    def encode_private_key(self) -> bytes | None: ...
    @abstractmethod
    def encode_public_key(self) -> bytes | None: ...
    @abstractmethod
    def sign(self, data: bytes) -> bytes: ...
    @abstractmethod
    def sign_verify(self, data: bytes, signature: bytes) -> bool: ...


class DILITHIUM(AsymCipher):
    keypair: _ty.Type[_DILITHIUM_KEYPAIRTYPE] = _DILITHIUM_KEYPAIRTYPE


_SPHINCS_MODES = ("sphincs-sha256-128s", "sphincs-sha256-192s", "sphincs-sha256-256s")
_SPHINCS_MODELITERAL = _ty.Literal[
    "sphincs-sha256-128s", "sphincs-sha256-192s", "sphincs-sha256-256s"
]


class _SPHINCS_KEYPAIRTYPE(
    _BaseAsymmetricKeypair, AsymKeySignable, AsymKeySerializable, metaclass=ABCMeta
):
    cipher = "SPHINCS+"

    def __init__(self, mode: _SPHINCS_MODELITERAL) -> None: ...

    @classmethod
    def new(cls, mode: _SPHINCS_MODELITERAL) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(f"{cls.cipher} not supported by this backend")
        if mode not in _SPHINCS_MODES:
            raise ValueError(f"mode must be one of {_SPHINCS_MODES}, not '{mode}'")
        return cls(mode)


class SPHINCS(AsymCipher):
    keypair: _ty.Type[_SPHINCS_KEYPAIRTYPE] = _SPHINCS_KEYPAIRTYPE


_FRODOKEM_MODES = ("FrodoKEM-640", "FrodoKEM-976", "FrodoKEM-1344")
_FRODOKEM_MODELITERAL = _ty.Literal["FrodoKEM-640", "FrodoKEM-976", "FrodoKEM-1344"]


class _FRODOKEM_KEYPAIRTYPE(
    _BaseAsymmetricKeypair, AsymEncapsulatable, AsymKeySerializable, metaclass=ABCMeta
):
    cipher = "FrodoKEM"

    def __init__(self, mode: _FRODOKEM_MODELITERAL) -> None: ...

    @classmethod
    def new(cls, mode: _FRODOKEM_MODELITERAL) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(f"{cls.cipher} not supported by this backend")
        if mode not in _FRODOKEM_MODES:
            raise ValueError(f"mode must be one of {_FRODOKEM_MODES}, not '{mode}'")
        return cls(mode)


class FRODOKEM(AsymCipher):
    keypair: _ty.Type[_FRODOKEM_KEYPAIRTYPE] = _FRODOKEM_KEYPAIRTYPE


_BIKE_MODES = ("bike1l1", "bike1l3", "bike1l5")
_BIKE_MODELITERAL = _ty.Literal["bike1l1", "bike1l3", "bike1l5"]


class _BIKE_KEYPAIRTYPE(
    _BaseAsymmetricKeypair, AsymEncapsulatable, AsymKeySerializable, metaclass=ABCMeta
):
    cipher = "BIKE"

    def __init__(self, mode: _BIKE_MODELITERAL) -> None: ...

    @classmethod
    def new(cls, mode: _BIKE_MODELITERAL) -> _te.Self:
        if not cls.__concrete__:
            raise NotSupportedError(f"{cls.cipher} not supported by this backend")
        if mode not in _BIKE_MODES:
            raise ValueError(f"mode must be one of {_BIKE_MODES}, not '{mode}'")
        return cls(mode)


class BIKE(AsymCipher):
    keypair: _ty.Type[_BIKE_KEYPAIRTYPE] = _BIKE_KEYPAIRTYPE



"""TBA"""
from importlib.resources import files
import unicodedata
import secrets
import string
import random
import math

from .rand import Random, SystemRandom, WeightedFunctions  # , OSRandom, SecretsRandom

# Standard typing imports for aps
import typing_extensions as _te
import collections.abc as _a
import typing as _ty

if _ty.TYPE_CHECKING:
    import _typeshed as _tsh
import types as _ts

__all__ = ["PasswordFilter", "SimplePasswordGenerator", "PasswordGenerator", "SecurePasswordGenerator"]


class PasswordFilter:
    """Used to easily filter passwords based on a few simple concepts."""

    def __init__(
        self,
        exclude_chars: str = "",
        extra_chars: str = "",
        exclude_similar: bool = False,
    ):
        self.exclude_chars: str = exclude_chars
        self.extra_chars: str = extra_chars
        self.exclude_similar: bool = exclude_similar
        self.extra_chars_dict: dict[str, str] = self._classify_extra_chars(extra_chars)

    def _classify_extra_chars(self, extra_chars: str) -> dict[str, str]:
        # Classify each extra char into the correct character set
        classified_chars = {
            "letters": "",
            "numbers": "",
            "punctuations": "",
            "unicode": "",
        }

        for char in extra_chars:
            if char in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ":
                classified_chars["letters"] += char
            elif char in "0123456789":
                classified_chars["numbers"] += char
            elif char in "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~":
                classified_chars["punctuations"] += char
            else:
                classified_chars["unicode"] += char

        return classified_chars

    def apply(
        self,
        character_set: str,
        set_name: _ty.Literal["letters", "numbers", "punctuations", "unicode"],
    ) -> str:
        """
        Applies the filter to a specific character set.

        :param character_set: The charset
        :param set_name: Which charset is being filtered
        :return: The filtered charset
        """
        if self.exclude_similar:
            similar_chars = "il1Lo0O"
            character_set = "".join(
                ch for ch in character_set if ch not in similar_chars
            )
        if self.exclude_chars:
            character_set = "".join(
                ch for ch in character_set if ch not in self.exclude_chars
            )
        character_set += self.extra_chars_dict[set_name]
        return character_set

    def filter_word(self, word: str) -> str:
        """
        Filters one word.

        :param word: The word to be filtered
        :return: The filtered word
        """
        # Filter the word according to the current settings
        filtered_word = "".join(ch for ch in word if ch not in self.exclude_chars)
        if self.exclude_similar:
            similar_chars = "il1Lo0O"
            filtered_word = "".join(
                ch for ch in filtered_word if ch not in similar_chars
            )
        return filtered_word

    def will_filter(self, char: str) -> bool:
        """
        Checks if a specific char would be filtered.

        :param char: The char to be checked
        :return: If the char would be filtered
        """
        return self.filter_word(char) == ""


@_te.deprecated("This class is deprecated, please use PasswordGenerator or SecurePasswordGenerator instead.")
class SimplePasswordGenerator:
    """Used to quickly generate a lot of passwords in minimal time"""

    _rng = random
    characters: str = string.ascii_letters + string.digits
    debug: bool = False

    @classmethod
    def _debug_print(cls, *args: _ty.Any, **kwargs: _ty.Any) -> None:
        if cls.debug:
            print(*args, **kwargs)

    @classmethod
    def quick_secure_password(
        cls, password: str, passes: int = 3, expand: bool = True
    ) -> str:
        """
        Quicker version of SpecificPasswordGenerator.secure_password

        :param password: The password to be secured
        :param passes: How many iterations this should run for
        :param expand: If the password should be able to get larger
        :return: The secured password
        """
        cls._debug_print(f"Quick securing password: {password}")
        characters = string.ascii_letters + string.digits + string.punctuation

        for _ in range(passes):
            if expand:
                # Add a random character
                password += cls._rng.choice(characters)
            else:
                rng = cls._rng.randint(0, len(password) - 1)
                password = (
                    password[:rng] + cls._rng.choice(characters) + password[rng + 1 :]
                )

            # Switch a random character's case
            pos = cls._rng.randint(0, len(password) - 1)
            char = password[pos]
            if char.islower():
                char = char.upper()
            elif char.isupper():
                char = char.lower()
            password = password[:pos] + char + password[pos + 1 :]

            # Add a random digit or punctuation
            pos = cls._rng.randint(0, len(password) - 1)
            password = (
                password[:pos]
                + cls._rng.choice(string.digits + string.punctuation)
                + password[(pos if expand else pos + 1) :]
            )

        # Shuffle the resulting password to ensure randomness
        password_list = list(password)
        cls._rng.shuffle(password_list)
        password = "".join(password_list)

        cls._debug_print(f"Quick secured password: {password}")
        return password

    @classmethod
    def generate_password(
        cls, length: int = 12, filter_: PasswordFilter | None = None
    ) -> str:
        """
        Entirely random (uses the random module)

        :param length: The length of the password
        :param filter_: An optional password filter
        :return: The generated password
        """
        cls._debug_print(f"Generating password of length {length}")
        characters = string.ascii_letters + string.digits + string.punctuation
        if filter_:
            characters = filter_.apply(characters, "letters")
            characters += filter_.apply(string.digits, "numbers")
            characters += filter_.apply(string.punctuation, "punctuations")
            cls._debug_print(f"Filtered characters: {characters}")
        password = "".join(cls._rng.choice(characters) for _ in range(length))
        cls._debug_print(f"Generated password: {password}")
        return password

    @classmethod
    def generate_secure_password(
        cls, length: int = 12, filter_: PasswordFilter | None = None
    ) -> str:
        """
        Entirely random (uses the secrets module)

        :param length: The length of the password
        :param filter_: An optional password filter
        :return: The generated password
        """
        cls._debug_print(f"Generating secure password of length {length}")
        characters = string.ascii_letters + string.digits + string.punctuation
        if filter_:
            characters = filter_.apply(characters, "letters")
            characters += filter_.apply(string.digits, "numbers")
            characters += filter_.apply(string.punctuation, "punctuations")
            cls._debug_print(f"Filtered characters: {characters}")
        password = "".join(secrets.choice(characters) for _ in range(length))
        cls._debug_print(f"Generated secure password: {password}")
        return password

    @classmethod
    def generate_passphrase(
        cls, words: list[str], num_words: int = 4, filter_: PasswordFilter | None = None
    ) -> str:
        """
        Passphrase, secure as it's very long

        :param words: Which words to use in the passphrase
        :param num_words: The number of words to be used
        :param filter_: An optional password filter
        :return: The generated password
        """
        cls._debug_print(f"Generating passphrase with {num_words} words")
        if filter_:
            words = list(filter(filter_.filter_word, words))
            cls._debug_print(f"Filtered words: {words[:10]}... (showing first 10)")
        passphrase = " ".join(cls._rng.choice(words) for _ in range(num_words))
        cls._debug_print(f"Generated passphrase: {passphrase}")
        return passphrase

    @classmethod
    def generate_pattern_password(
        cls, pattern: str = "XXX-999-xxx", filter_: PasswordFilter | None = None
    ) -> str:
        """
        A pattern password X for an uppercase letter, x for a lowercase letter and 9 for a number.
        All other characters are ignored and included as is in the final password

        :param pattern: The pattern to generate the password by
        :param filter_: An optional password filter
        :return: The generated password
        """
        cls._debug_print(f"Generating pattern password with pattern {pattern}")

        def _random_char(c: str) -> str:
            if c == "X":
                chars = string.ascii_uppercase
            elif c == "x":
                chars = string.ascii_lowercase
            elif c == "9":
                chars = string.digits
            else:
                return c

            if filter_:
                chars = filter_.apply(
                    chars,
                    "letters"
                    if c in "Xx"
                    else "numbers"
                    if c == "9"
                    else "punctuations",
                )
                cls._debug_print(f"Filtered characters for '{c}': {chars}")

            return str(cls._rng.choice(chars))

        password = "".join(_random_char(c) for c in pattern)
        cls._debug_print(f"Generated pattern password: {password}")
        return password

    @classmethod
    def generate_complex_password(
        cls, length: int = 12, filter_: PasswordFilter | None = None
    ) -> str:
        """
        Entirely random, a complex password.

        :param length: The length of the password
        :param filter_: An optional password filter
        :return: The generated password
        """
        cls._debug_print(f"Generating complex password of length {length}")
        if length < 4:
            raise ValueError("Password length should be at least 4")

        all_characters = string.ascii_letters + string.digits + string.punctuation
        if filter_:
            all_characters = filter_.apply(all_characters, "letters")
            all_characters += filter_.apply(string.digits, "numbers")
            all_characters += filter_.apply(string.punctuation, "punctuations")
            cls._debug_print(f"Filtered all characters: {all_characters}")

        if filter_:
            password = [
                cls._rng.choice(filter_.apply(string.ascii_uppercase, "letters")),
                cls._rng.choice(filter_.apply(string.ascii_lowercase, "letters")),
                cls._rng.choice(filter_.apply(string.digits, "numbers")),
                cls._rng.choice(filter_.apply(string.punctuation, "punctuations")),
            ]
        else:
            password = [
                cls._rng.choice(string.ascii_uppercase),
                cls._rng.choice(string.ascii_lowercase),
                cls._rng.choice(string.digits),
                cls._rng.choice(string.punctuation),
            ]

        remaining_chars = all_characters
        password += [cls._rng.choice(remaining_chars) for _ in range(length - 4)]
        cls._rng.shuffle(password)

        cls._debug_print(f"Generated complex password: {''.join(password)}")
        return "".join(password)

    @classmethod
    def generate_mnemonic_password(cls, filter_: PasswordFilter | None = None) -> str:
        """
        A mnemonic password using adjectives and nouns

        :param filter_: An optional password filter
        :return:
        """
        cls._debug_print("Generating mnemonic password")
        adjectives = ["quick", "lazy", "sleepy", "noisy", "hungry"]
        nouns = ["fox", "dog", "cat", "mouse", "bear"]
        symbols = string.punctuation
        numbers = string.digits

        adj = cls._rng.choice(adjectives)
        noun = cls._rng.choice(nouns)
        symbol = cls._rng.choice(symbols)
        number = cls._rng.choice(numbers)

        if filter_:
            adj = filter_.filter_word(adj)
            noun = filter_.filter_word(noun)
            symbol = cls._rng.choice(filter_.apply(symbols, "punctuations"))
            number = cls._rng.choice(filter_.apply(numbers, "numbers"))

        password = f"{adj}{symbol}{noun}{number}"
        cls._debug_print(f"Generated mnemonic password: {password}")
        return password

    @classmethod
    def generate_ratio_based_password(
        cls,
        length: int = 12,
        letter_ratio: float = 0.4,
        digit_ratio: float = 0.3,
        symbol_ratio: float = 0.3,
        filter_: PasswordFilter | None = None,
    ) -> str:
        """
        Generate a simple ratio based password

        :param length: The length of the password
        :param letter_ratio: The ratio of letters in the final password
        :param digit_ratio: The ratio of digits in the final password
        :param symbol_ratio: The ratio of symbols in the final password
        :param filter_: An optional password filter
        :return: The generated password
        """
        cls._debug_print(f"Generating ratio-based password of length {length}")

        num_letters = int(length * letter_ratio)
        num_digits = int(length * digit_ratio)
        num_symbols = length - num_letters - num_digits

        characters = (
            (
                filter_.apply(string.ascii_letters, "letters")
                if filter_
                else string.ascii_letters
            )
            * num_letters
            + (filter_.apply(string.digits, "numbers") if filter_ else string.digits)
            * num_digits
            + (
                filter_.apply(string.punctuation, "punctuations")
                if filter_
                else string.punctuation
            )
            * num_symbols
        )

        password = "".join(cls._rng.choice(characters) for _ in range(length))
        cls._debug_print(f"Generated ratio-based password: {password}")
        return password

    @classmethod
    def generate_sentence_based_password(cls, structure: str = "WwWn!") -> str:
        """
        A sentence based password, using a structure.
        W means a capitalized word, w means a lowercase word, n means a digit and ! means a symbol.

        :param structure:
        :return:
        """
        cls._debug_print(
            f"Generating sentence-based password with structure {structure}"
        )

        words = ["quick", "brown", "fox", "jumps", "over", "lazy", "dog"]
        symbols = string.punctuation
        digits = string.digits

        password = []
        for char in structure:
            if char == "W":
                password.append(cls._rng.choice(words).capitalize())
            elif char == "w":
                password.append(cls._rng.choice(words))
            elif char == "n":
                password.append(cls._rng.choice(digits))
            elif char == "!":
                password.append(cls._rng.choice(symbols))

        password_str = "".join(password)
        cls._debug_print(f"Generated sentence-based password: {password_str}")
        return password_str

    @classmethod
    def generate_custom_sentence_based_password_v2(
        cls,
        sentence: str,
        char_position: _ty.Literal["random", "keep"] | int = "random",
        random_case: bool = False,
        extra_char: str = "",
        num_length: int = 0,
        special_chars_length: int = 0,
    ) -> str:
        """
        A sentence based password, v2.

        :param sentence: The sentence to generate the password with, if this is not passed, a random sentence will be used.
        :param char_position: Which character in the word should be used in the password. You can also pass "random" for random and "keep" for keeping their position.
        :param random_case: If the case of characters in the final password should be randomized
        :param extra_char: Extra character part of the password
        :param num_length: The length of the numbers part of the password
        :param special_chars_length: The length of the punctuation part of the password
        :return: The generated password
        """
        words = sentence.split(" ")
        word_chars = []
        for word in words:
            if char_position == "random":
                index = cls._rng.randint(0, len(word) - 1)
            elif char_position == "keep":
                index = 0
            elif type(char_position) is int:
                index = min(char_position, len(word) - 1)
            else:
                return "Invalid char_position."
            char = word[index]
            if random_case:
                char = char.lower() if cls._rng.random() < 0.5 else char.upper()
            word_chars.append(char)
            cls._debug_print("Word characters after processing:", word_chars)
        num_string = "".join(cls._rng.choices("0123456789", k=num_length))
        cls._debug_print("Numeric string after generation:", num_string)
        special_chars_string = "".join(
            cls._rng.choices(
                "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~", k=special_chars_length
            )
        )
        cls._debug_print(
            "Special characters string after generation:", special_chars_string
        )
        password = "".join(word_chars) + extra_char + num_string + special_chars_string
        cls._debug_print("Final password:", password)
        return password


class PasswordGenerator:
    """The random generators security can be adjusted, and all default arguments will result in a password
    that has an average worst case crack time of multiple years (Verified using zxcvbn)."""

    def __init__(self, random_generator: Random, debug: bool = False) -> None:
        self._rng = random_generator
        self.debug: bool = debug
        self.words: list[str] = []

    def _debug_print(self, *args: _ty.Any, **kwargs: _ty.Any) -> None:
        if self.debug:
            print(*args, **kwargs)

    def _add_unique_words(self, words: list[str]) -> None:
        unique_words = set(
            self.words
        )  # Create a set of current words for fast membership testing
        for word in words:
            if word not in unique_words:
                self.words.append(word)
                unique_words.add(word)

    def load_def_dict(self) -> None:
        """
        Loads the default dictionary for random sentence generation.
        :return:
        """
        from pathlib import Path
        file_path = files("acecurity.dicts").joinpath("def-dict.txt")
        with file_path.open("r") as f:
            self._add_unique_words([line.strip() for line in f])

    def load_google_10000_dict(self) -> None:
        """
        Loads google 10 000 dictionary for random sentence generation (uk & us without swears).
        :return:
        """
        from pathlib import Path
        file_path = files("acecurity.dicts").joinpath("google-10000-dict.txt")
        with file_path.open("r") as f:
            self._add_unique_words([line.strip() for line in f])

    def load_scowl_dict(
        self, size: _ty.Literal["50", "60", "70", "80", "95"] = "50"
    ) -> None:
        """
        Loads one of scowl dicts, specified by their size.
        Larger sizes have all words of the smaller ones and more.
        :param size: The size of the dictionary to be loaded
        :type size: Literal["50", "60", "70", "80", "95"]
        :return:
        """
        from pathlib import Path
        file_path = files("acecurity.dicts").joinpath(f"scowl-{size}-dict.txt")
        with file_path.open("r") as f:
            self._add_unique_words([line.strip() for line in f])

    def load_12_dicts(self) -> None:
        """
        Loads a selection of the 12 dicts which were deemed the most useful in sentence generation.
        :return:
        """
        from pathlib import Path
        dir_path = files("acecurity.dicts").joinpath("12-dicts")
        for file in dir_path.iterdir():
            with file.open("r") as f:
                self._add_unique_words([line.strip() for line in f])

    def unload_dicts(self) -> None:
        """Unloads all currently loaded dictionaries."""
        self.words.clear()

    def basic_secure_password(
        self, password: str, passes: int = 3, expand: bool = True
    ) -> str:
        """
        Makes the inputted password more secure by introducing slight changes each iteration.

        :param password: The password to be secured
        :param passes: How many iterations this should run for
        :param expand: If the password should be able to get larger
        :return: The secured password
        """
        self._debug_print(f"Securing password: {password}")
        characters = string.ascii_letters + string.digits + string.punctuation

        for _ in range(passes):
            rnd = self._rng.randint(0, 3)

            match rnd:
                case 0:
                    if expand:
                        # Add a random character
                        password += self._rng.choice(characters)
                    else:
                        rng = self._rng.randint(0, len(password) - 1)
                        password = (
                            password[:rng]
                            + self._rng.choice(characters)
                            + password[rng + 1 :]
                        )
                case 1:
                    # Switch a random character's case
                    pos = self._rng.randint(0, len(password) - 1)
                    char = password[pos]
                    if char.islower():
                        char = char.upper()
                    elif char.isupper():
                        char = char.lower()
                    password = password[:pos] + char + password[pos + 1 :]
                case 2:
                    # Add a random digit
                    pos = self._rng.randint(0, len(password) - 1)
                    password = (
                        password[:pos]
                        + self._rng.choice(string.digits)
                        + password[(pos if expand else pos + 1) :]
                    )
                case 3:
                    # Add a random punctuation
                    pos = self._rng.randint(0, len(password) - 1)
                    password = (
                        password[:pos]
                        + self._rng.choice(string.punctuation)
                        + password[(pos if expand else pos + 1) :]
                    )

        # Calculate the number of characters to switch
        length = len(password)
        percentage_to_switch = self._rng.choice([10, 20, 30])
        num_to_switch = max(2, (length * percentage_to_switch) // 100)

        # Switch around the selected percentage of characters
        password_list = list(password)
        for _ in range(num_to_switch // 2):
            idx1, idx2 = self._rng.sample(list(range(length)), 2)
            password_list[idx1], password_list[idx2] = (
                password_list[idx2],
                password_list[idx1],
            )

        self._debug_print(f"Secured password: {password}")
        return password

    def dynamic_secure_password(
        self, password: str, passes: int = 3, expand: bool = True
    ) -> str:
        """
        Makes the inputted password more secure by introducing slight changes each iteration.
        Dynamically changes less the longer the password, as they are already really secure.

        :param password: The password to be secured
        :param passes: How many iterations this should run for
        :param expand: If the password should be able to get larger
        :return: The secured password
        """
        self._debug_print(f"Securing password: {password}")
        characters = string.ascii_letters + string.digits + string.punctuation

        for _ in range(passes):
            if expand:
                # Add a random character
                password += self._rng.choice(characters)
            else:
                rng = self._rng.randint(0, len(password) - 1)
                password = (
                    password[:rng] + self._rng.choice(characters) + password[rng + 1 :]
                )

            # Switch a random character's case
            pos = self._rng.randint(0, len(password) - 1)
            char = password[pos]
            if char.islower():
                char = char.upper()
            elif char.isupper():
                char = char.lower()
            password = password[:pos] + char + password[pos + 1 :]

            # Add a random digit
            pos = self._rng.randint(0, len(password) - 1)
            password = (
                password[:pos]
                + self._rng.choice(string.digits)
                + password[(pos if expand else pos + 1) :]
            )

            # Add a random punctuation
            pos = self._rng.randint(0, len(password) - 1)
            password = (
                password[:pos]
                + self._rng.choice(string.punctuation)
                + password[(pos if expand else pos + 1) :]
            )

        # Calculate the number of characters to switch
        length = len(password)
        percentage_to_switch = self._rng.choice([10, 20, 30])
        num_to_switch = max(2, (length * percentage_to_switch) // 100)

        # Switch around the selected percentage of characters
        password_list = list(password)
        for _ in range(num_to_switch // 2):
            idx1, idx2 = self._rng.sample(list(range(length)), 2)
            password_list[idx1], password_list[idx2] = (
                password_list[idx2],
                password_list[idx1],
            )

        self._debug_print(f"Secured password: {password}")
        return password

    def static_secure_password(
        self, password: str, passes: int = 3, expand: bool = True
    ) -> str:
        """
        Makes the inputted password more secure by introducing slight changes each iteration.

        :param password: The password to be secured
        :param passes: How many iterations this should run for
        :param expand: If the password should be able to get larger
        :return: The secured password
        """
        self._debug_print(f"Securing password: {password}")
        characters = string.ascii_letters + string.digits + string.punctuation
        length = len(password)

        def get_num() -> int:
            return max(2, (length * self._rng.choice([10, 20, 30])) // 100)

        for _ in range(passes):
            for _ in range(get_num()):
                if expand:
                    # Add a random character
                    password += self._rng.choice(characters)
                else:
                    rng = self._rng.randint(0, len(password) - 1)
                    password = (
                        password[:rng]
                        + self._rng.choice(characters)
                        + password[rng + 1 :]
                    )

                # Switch a random character's case
                pos = self._rng.randint(0, len(password) - 1)
                char = password[pos]
                if char.islower():
                    char = char.upper()
                elif char.isupper():
                    char = char.lower()
                password = password[:pos] + char + password[pos + 1 :]

                # Add a random digit
                pos = self._rng.randint(0, len(password) - 1)
                password = (
                    password[:pos]
                    + self._rng.choice(string.digits)
                    + password[(pos if expand else pos + 1) :]
                )

                # Add a random punctuation
                pos = self._rng.randint(0, len(password) - 1)
                password = (
                    password[:pos]
                    + self._rng.choice(string.punctuation)
                    + password[(pos if expand else pos + 1) :]
                )

        # Calculate the number of characters to switch
        num_to_switch = get_num()

        # Switch around the selected percentage of characters
        password_list = list(password)
        for _ in range(num_to_switch // 2):
            idx1, idx2 = self._rng.sample(list(range(length)), 2)
            password_list[idx1], password_list[idx2] = (
                password_list[idx2],
                password_list[idx1],
            )

        self._debug_print(f"Secured password: {password}")
        return password

    def generate_secure_sentence(self, length: int = 4) -> str:
        """
        Generates a secure sentence using the currently loaded dictionaries,
        often used internally for secure sentence generation.
        [Choices is more secure than sample]

        :param length: The length of the sentence in words
        :return: The generated sentence
        """
        return " ".join(self._rng.choices(self.words, k=length))

    def reduce_password(self, password: str, by: int | _ty.Literal["all"] = 0) -> str:
        """
        Reduces a given password using random binary operations to ensure a secure password.

        :param password: The password to be reduces
        :param by: How many characters each character should be joined with (0 means none and "all" means all).
        :return: The reduced password
        """
        byi: int = int(by if by != "all" else len(password))
        result = []

        for chunk in (
            password[i : i + byi + 1] for i in range(0, len(password), byi + 1)
        ):
            if byi == 0:
                chunk = (
                    chunk + chunk
                )  # Duplicate the character to make a two-character chunk
            while len(chunk) < byi + 1:
                chunk += chunk  # Repeat the chunk until it reaches the desired size
            chunk = chunk[: byi + 1]  # Trim the chunk to the exact desired size

            rnd = self._rng.randint(0, 5)
            reduced_char = chunk[0]

            match rnd:
                case 0:
                    for char in chunk[1:]:
                        reduced_char = chr(ord(reduced_char) | ord(char))
                case 1:
                    for char in chunk[1:]:
                        reduced_char = chr(ord(reduced_char) & ord(char))
                case 2:
                    for char in chunk[1:]:
                        reduced_char = chr(ord(reduced_char) ^ ord(char))
                case 3:
                    reduced_char = chr(~ord(reduced_char) & 0xFF)  # NOT operation
                case 4:
                    for char in chunk[1:]:
                        reduced_char = chr(
                            (ord(reduced_char) << 1) & 0xFF
                        )  # Left shift
                case 5:
                    for char in chunk[1:]:
                        reduced_char = chr(ord(reduced_char) >> 1)  # Right shift

            result.append(reduced_char)

        return "".join(result)

    def generate_random_password(
        self, length: int = 18, filter_: PasswordFilter | None = None
    ) -> str:
        """
        Entirely random

        :param length: The length of the password
        :param filter_: An optional password filter
        :return: The generated password
        """
        self._debug_print(f"Generating password of length {length}")
        characters = string.ascii_letters + string.digits + string.punctuation
        if filter_:
            characters = filter_.apply(characters, "letters")
            characters += filter_.apply(string.digits, "numbers")
            characters += filter_.apply(string.punctuation, "punctuations")
            self._debug_print(f"Filtered characters: {characters}")
        password = "".join(self._rng.choice(characters) for _ in range(length))
        self._debug_print(f"Generated password: {password}")
        return password

    def _generate_passphrase(
        self,
        words: list[str] | None = None,
        num_words: int = 6,
        filter_: PasswordFilter | None = None,
    ) -> str:
        if words is None:
            words = [
                "These",
                "can",
                "be",
                "insecure",
                "as",
                "long",
                "as",
                "the",
                "password",
                "is",
                "long",
            ]
        self._debug_print(f"Generating passphrase with {num_words} words")
        if filter_:
            words = list(filter(filter_.filter_word, words))
            self._debug_print(f"Filtered words: {words[:10]}... (showing first 10)")
        passphrase = " ".join(self._rng.choice(words) for _ in range(num_words))
        self._debug_print(f"Generated passphrase: {passphrase}")
        return passphrase

    def generate_secure_passphrase(
        self,
        words: list[str] | None = None,
        num_words: int = 6,
        filter_: PasswordFilter | None = None,
        _return_info: bool = True,
    ) -> str | tuple[str, str]:
        """
        Passphrase, secure as it's very long

        :param words: Which words to use in the passphrase
        :param num_words: The number of words to be used
        :param filter_: An optional password filter
        :param _return_info: If extra info about the password should be returned
        :return: The generated password
        """
        passphrase: str = self._generate_passphrase(words, num_words, filter_)
        secure_passphrase: str = self.dynamic_secure_password(passphrase, passes=1)
        return (secure_passphrase, passphrase) if _return_info else secure_passphrase

    def generate_pattern_password(
        self,
        pattern: str = "9/XxX-999xXx-99/XxX-999xXx-9",
        filter_: PasswordFilter | None = None,
    ) -> str:
        """
        A pattern password X for an uppercase letter, x for a lowercase letter and 9 for a number.
        All other characters are ignored and included as is in the final password

        :param pattern: The pattern to generate the password by
        :param filter_: An optional password filter
        :return: The generated password
        """
        self._debug_print(f"Generating pattern password with pattern {pattern}")
        if filter_:
            upper = filter_.apply(string.ascii_uppercase, "letters")
            self._debug_print(
                f"Filtered characters for '{string.ascii_uppercase}': {upper}"
            )
            lower = filter_.apply(string.ascii_lowercase, "letters")
            self._debug_print(
                f"Filtered characters for '{string.ascii_lowercase}': {lower}"
            )
            digits = filter_.apply(string.digits, "numbers")
            self._debug_print(f"Filtered characters for '{string.digits}': {digits}")
        else:
            upper = string.ascii_uppercase
            lower = string.ascii_lowercase
            digits = string.digits

        def _random_char(c: str) -> str:
            if c == "X":
                chars = upper
            elif c == "x":
                chars = lower
            elif c == "9":
                chars = digits
            else:
                return c

            return self._rng.choice(chars)

        password = "".join(_random_char(c) for c in pattern)
        self._debug_print(f"Generated pattern password: {password}")
        return password

    def generate_complex_pattern_password(
        self,
        pattern: str = "9nX9XWn9Ww9x9X-W!/999wnwwx!!",
        filter_: PasswordFilter | None = None,
    ) -> str:
        """
        A pattern password X for an uppercase letter, x for a lowercase letter, 9 for a number,
        W for a capitalized word, w for a lowercase word, n for a negative number and ! for a symbol.
        All other characters are ignored and included as is in the final password

        :param pattern: The pattern to generate the password by
        :param filter_: An optional password filter
        :return: The generated password
        """
        self._debug_print(f"Generating pattern password with pattern {pattern}")
        if filter_:
            upper = filter_.apply(string.ascii_uppercase, "letters")
            self._debug_print(
                f"Filtered characters for '{string.ascii_uppercase}': {upper}"
            )
            lower = filter_.apply(string.ascii_lowercase, "letters")
            self._debug_print(
                f"Filtered characters for '{string.ascii_lowercase}': {lower}"
            )
            digits = filter_.apply(string.digits, "numbers")
            self._debug_print(f"Filtered characters for '{string.digits}': {digits}")
            symbols = filter_.apply(string.punctuation, "punctuations")
            self._debug_print(
                f"Filtered characters for '{string.punctuation}': {digits}"
            )
        else:
            upper = string.ascii_uppercase
            lower = string.ascii_lowercase
            digits = string.digits
            symbols = string.punctuation
        password = []

        for char in pattern:
            match char:
                case "X":
                    password.append(self._rng.choice(upper))
                case "x":
                    password.append(self._rng.choice(lower))
                case "9":
                    password.append(self._rng.choice(digits))
                case "W":
                    word = self._rng.choice(self.words)
                    if filter_:
                        word = filter_.filter_word(word)
                    password.append(word.title())
                case "w":
                    word = self._rng.choice(self.words)
                    if filter_:
                        word = filter_.filter_word(word)
                    password.append(word)
                case "n":
                    password.append("-" + self._rng.choice(digits))
                case "!":
                    password.append(self._rng.choice(symbols))
                case _:
                    password.append(char)

        return "".join(password)

    def generate_complex_password(
        self, length: int = 18, filter_: PasswordFilter | None = None
    ) -> str:
        """
        Generate a complex, high-entropy password using random characters.

        The password includes at least one uppercase letter, one lowercase letter,
        one digit, and one punctuation character. The remaining characters are selected
        randomly from all allowed character sets. A password filter can optionally
        constrain which characters are used.

        :param length: The length of the password
        :param filter_: An optional password filter
        :return: The generated password
        """
        self._debug_print(f"Generating complex password of length {length}")
        if length < 4:
            raise ValueError("Password length should be at least 4")

        all_characters = string.ascii_letters + string.digits + string.punctuation

        if filter_:
            upper_chars = filter_.apply(string.ascii_uppercase, "letters")
            lower_chars = filter_.apply(string.ascii_lowercase, "letters")
            digit_chars = filter_.apply(string.digits, "numbers")
            punct_chars = filter_.apply(string.punctuation, "punctuations")

            all_characters = (
                filter_.apply(all_characters, "letters")
                + filter_.apply(string.digits, "numbers")
                + filter_.apply(string.punctuation, "punctuations")
            )
        else:
            upper_chars = string.ascii_uppercase
            lower_chars = string.ascii_lowercase
            digit_chars = string.digits
            punct_chars = string.punctuation

        self._debug_print(f"Filtered all characters: {all_characters}")

        password = [
            self._rng.choice(upper_chars),
            self._rng.choice(lower_chars),
            self._rng.choice(digit_chars),
            self._rng.choice(punct_chars),
        ]

        remaining_chars = all_characters
        password += [self._rng.choice(remaining_chars) for _ in range(length - 4)]
        self._rng.shuffle(password)

        self._debug_print(f"Generated complex password: {''.join(password)}")
        return "".join(password)

    def generate_mnemonic_password(
        self,
        length: int = 20,
        filter_: PasswordFilter | None = None,
        _return_info: bool = False,
    ) -> str | tuple[str, str, str]:
        """
        A mnemonic password using adjectives and nouns

        :param length: The length of the password
        :param filter_: An optional password filter
        :param _return_info: If extra info about the password should be returned
        :return:
        """
        self._debug_print("Generating mnemonic password")
        adjectives = [
            "quick",
            "lazy",
            "sleepy",
            "noisy",
            "hungry",
            "brave",
            "clever",
            "fierce",
            "jolly",
            "kind",
        ]
        nouns = [
            "fox",
            "dog",
            "cat",
            "mouse",
            "bear",
            "tiger",
            "lion",
            "wolf",
            "eagle",
            "shark",
        ]
        symbols = string.punctuation
        numbers = string.digits

        adj = self._rng.choice(adjectives)
        noun = self._rng.choice(nouns)

        if filter_:
            adj = filter_.filter_word(adj)
            noun = filter_.filter_word(noun)

        # Randomize case
        adj = "".join(
            [char.upper() if self._rng.random() > 0.5 else char for char in adj]
        )
        noun = "".join(
            [char.upper() if self._rng.random() > 0.5 else char for char in noun]
        )

        # Calculate additional characters needed
        extra_length = max(0, length - (len(adj) + len(noun)))
        symbols_needed = extra_length // 2
        numbers_needed = extra_length - symbols_needed

        extra_symbols = "".join(
            self._rng.choice(
                filter_.apply(symbols, "punctuations") if filter_ else symbols
            )
            for _ in range(symbols_needed)
        )
        extra_numbers = "".join(
            self._rng.choice(filter_.apply(numbers, "numbers") if filter_ else numbers)
            for _ in range(numbers_needed)
        )

        password = f"{adj}{extra_symbols}{noun}{extra_numbers}"
        self._debug_print(f"Generated mnemonic password: {password}")
        return (password, adj, noun) if _return_info else password

    def generate_ratio_based_password_v3(
        self,
        length: int = 24,
        letters_ratio: float = 0.5,
        numbers_ratio: float = 0.3,
        punctuations_ratio: float = 0.2,
        unicode_ratio: float = 0.0,
        filter_: PasswordFilter | None = None,
    ) -> str:
        """
        Generate a version 3 password based on character type ratios.

        This method creates a password composed of randomly selected characters,
        distributed according to specified ratios of letters, numbers, punctuation,
        and optional Unicode symbols.

        :param length: The length of the password
        :param letters_ratio: The ratio of letters in the final password
        :param numbers_ratio: The ratio of numbers in the final password
        :param punctuations_ratio: The ratio of punctuations in the final password
        :param unicode_ratio: The ratio of unicode in the final password
        :param filter_: An optional password filter
        :return:
        """
        character_sets: dict[str, str] = {
            "letters": "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
            "numbers": "0123456789",
            "punctuations": "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~",
        }
        if unicode_ratio != 0.0:
            character_sets["unicode"] = "".join(
                chr(i)
                for i in range(0x110000)
                if unicodedata.category(chr(i)).startswith("P")
            )
        self._debug_print(f"Generated char_sets {character_sets}")

        if filter_:
            for key in character_sets:
                if character_sets[key]:
                    character_sets[key] = filter_.apply(character_sets[key], key)  # type: ignore

        self._debug_print(f"Filtered char_sets {character_sets}")

        non_zero_lengths = {
            ratio_name: ratio
            for ratio_name, ratio in {
                "letters": math.ceil(length * letters_ratio),
                "numbers": math.ceil(length * numbers_ratio),
                "punctuations": math.ceil(length * punctuations_ratio),
                "unicode": math.ceil(length * unicode_ratio),
            }.items()
            if ratio != 0 and len(character_sets[ratio_name]) != 0
        }
        remaining_length = length - sum(non_zero_lengths.values())

        self._debug_print(
            f"Generated non-zero lengths {non_zero_lengths}, remaining_length {remaining_length}"
        )
        if remaining_length != 0:
            keys = list(non_zero_lengths.keys())
            for i in range(abs(remaining_length)):
                non_zero_lengths[keys[i % len(non_zero_lengths)]] += (
                    1 if remaining_length > 0 else -1
                )
        self._debug_print(
            f"After adjustment to chosen length, lengths {non_zero_lengths}"
        )

        # Generate password characters
        password_characters: list[str] = []
        for key, num_chars in non_zero_lengths.items():
            if character_sets[key]:
                password_characters.extend(
                    self._rng.choices(character_sets[key], k=num_chars)
                )

        self._debug_print(f"Generated password chars {password_characters}")

        # Shuffle the resulting password characters to ensure randomness
        self._rng.shuffle(password_characters)

        self._debug_print(f"Final password chars {password_characters}")

        return "".join(password_characters)

    def generate_words_based_password_v3(
        self,
        sentence: str | None = None,
        shuffle_words: bool = True,
        shuffle_characters: bool = True,
        repeat_words: bool = False,
        filter_: PasswordFilter | None = None,
        _return_info: bool = False,
    ) -> str | tuple[str, str]:
        """
        Generate a version 3 words-based password.

        This method creates a password using words from a sentence. It allows customization
        such as shuffling the word order, scrambling characters within words, and repeating words.
        If no sentence is provided, a random one is used.

        :param sentence: The sentence to generate the password with, if this is not passed, a random sentence will be used.
        :param shuffle_words: If the order of the words should change in the final password
        :param shuffle_characters: If the order of the characters in each word should change
        :param repeat_words: If words should be able to be repeated
        :param filter_: An optional password filter
        :param _return_info: If extra info about the password should be returned
        :return: The generated password
        """
        if sentence is None:
            sentence = self.generate_secure_sentence(length=5)
        words = sentence.split(" ")

        self._debug_print("Words", words)
        if shuffle_words:
            self._rng.shuffle(words)
            self._debug_print("Words after shuffling", words)

        if shuffle_characters:
            words = ["".join(self._rng.sample(list(word), len(word))) for word in words]
            self._debug_print("Words after character shuffling", words)

        if filter_:
            words = [filter_.filter_word(word) for word in words]

        if repeat_words:
            length = len(words)
            password = "".join(self._rng.choice(words) for _ in range(length))
        else:
            password = "".join(words)

        self._debug_print("Generated password", password)
        return (password, sentence) if _return_info else password

    def generate_sentence_based_password_v3(
        self,
        sentence: str | None = None,
        char_position: _ty.Literal["random", "keep"] | int = "keep",
        random_case: bool = True,
        extra_char: str = "/",
        num_length: int = 6,
        special_chars_length: int = 4,
        password_format: str = "{words}{special}{extra}{numbers}",
        filter_: PasswordFilter | None = None,
        _return_info: bool = False,
    ) -> str | tuple[str, str]:
        """
        Generate a version 3 sentence-based password.

        This method creates a password using words from a sentence (either user-provided or randomly generated),
        combined with random numbers, special characters, and an optional extra character. The format and rules
        for how the password is constructed can be customized.

        :param sentence: The sentence to generate the password with, if this is not passed, a random sentence will be used.
        :param char_position: Which character in the word should be used in the password. You can also pass "random" for random and "keep" for keeping their position.
        :param random_case: If the case of characters in the final password should be randomized
        :param extra_char: Extra character part of the password
        :param num_length: The length of the numbers part of the password
        :param special_chars_length: The length of the punctuation part of the password
        :param password_format: How to structure the different parts of the password into one.
        :param filter_: An optional password filter
        :param _return_info: If extra info about the password should be returned
        :return: The generated password
        """
        if sentence is None:
            sentence = self.generate_secure_sentence(length=8)
        words = sentence.split(" ")
        word_chars = []
        for word in words:
            if not word:
                continue

            if char_position == "random":
                indices = list(range(len(word)))
                self._rng.shuffle(indices)
            elif char_position == "keep":
                indices = [x for x in range(len(word))]
            elif isinstance(char_position, int):
                indices = [min(char_position, len(word) - 1)]
            else:
                return "Invalid char_position."

            char_selected = False
            for index in indices:
                char = word[index]
                if random_case:
                    char = char.lower() if self._rng.random() < 0.5 else char.upper()
                if filter_ is None or not filter_.will_filter(char):
                    word_chars.append(char)
                    self._debug_print("Word characters after processing:", word_chars)
                    char_selected = True
                    break

            if not char_selected:
                self._debug_print("No valid character found for word:", word)

        word_string = "".join(word_chars)
        num_string = "".join(self._rng.choices("0123456789", k=num_length))
        self._debug_print("Numeric string after generation:", num_string)
        special_chars_string = "".join(
            self._rng.choices(
                "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~", k=special_chars_length
            )
        )
        self._debug_print(
            "Special characters string after generation:", special_chars_string
        )

        password = password_format.format(
            words=word_string,
            extra=extra_char,
            numbers=num_string,
            special=special_chars_string,
        )
        self._debug_print("Final password:", password)

        return (password, sentence) if _return_info else password


class SecurePasswordGenerator:
    """
    Generates an always secure password that takes at least over one year to crack in the worst case
    (Verified by zxcvbn).
    This is also guaranteed if some of your passwords are already compromised as the password generation pattern is also
    entirely random.

    The number of unique words for random sentence generation grows with the security settings (All duplicates are
    filtered out):
    - Weak: 10.027 words (random module for random number generation)
    - Average: 105.705 words (numpy.random module for random number generation)
    - Strong: 223.242 words (secrets module for random number generation)
    - Super Strong: 400.172 words (secrets module for random number generation)
    """

    def __init__(
        self, security: _ty.Literal["weak", "average", "strong", "super_strong"]
    ):
        self._rng = {
            "weak": Random,
            "average": Random,
            "strong": SystemRandom,
            "super_strong": SystemRandom,
        }[security]()
        self._gen = PasswordGenerator(self._rng)

        match security:
            case "weak":
                self._gen.load_def_dict()
                self._gen.load_google_10000_dict()
            case "average":
                self._gen.load_def_dict()
                self._gen.load_google_10000_dict()
                self._gen.load_scowl_dict()
            case "strong":
                self._gen.load_def_dict()
                self._gen.load_google_10000_dict()
                self._gen.load_scowl_dict(size="70")
                self._gen.load_12_dicts()
            case "super_strong":
                self._gen.load_def_dict()
                self._gen.load_google_10000_dict()
                self._gen.load_scowl_dict(size="80")
                self._gen.load_12_dicts()
        print(f"Loaded {len(self._gen.words):,} unique words")

    def _exponential(
        self,
        lower_bound: float | int = 0,
        upper_bound: float | int = 1,
        multiplier: float = 1.0,
    ) -> float:
        """
        Generate a random number based on the exponential distribution and scale it within the specified bounds.

        :param lower_bound: Lower bound of the scaled range.
        :param upper_bound: Upper bound of the scaled range.
        :param multiplier: Multiplier parameter x**mult e of the distribution.
        :return: Scaled random number from the exponential distribution.
        """
        transformed_value: float = (lambda x: x ** (multiplier * math.e))(
            self._rng.random()
        )
        scaled_value: float = (
            lower_bound + (upper_bound - lower_bound) * transformed_value
        )
        return scaled_value

    def random(self) -> dict[str, str]:
        """
        Generate a fully random password with no specific pattern or structure.

        This method delegates to the underlying generator to create a password
        composed of random characters from a general character set, with no
        guarantees on format or composition.
        :return: The password and generation info.
        """
        pw = self._gen.generate_random_password()
        return {"extra_info": "Entirely random", "password": pw}

    def passphrase(self) -> dict[str, str]:
        """
        Generate a secure, long passphrase based on a sentence.

        This method creates a passphrase by selecting multiple words (typically from a wordlist)
        to form a human-readable but high-entropy sentence. It returns both the passphrase and
        an explanation of the sentence used to generate it.
        :return: The password and generation info.
        """
        pw, sentence = self._gen.generate_secure_passphrase(
            self._gen.words, _return_info=True
        )
        return {
            "extra_info": f"Passphrase, secure as it's very long, using the sentence '{sentence}'",
            "password": pw,
        }

    def pattern(self) -> dict[str, str]:
        """
        Generate a password based on a randomized character pattern.

        This method constructs a random password pattern by selecting 28 characters
        from the template set: `/`, `X` (uppercase), `x` (lowercase), `9` (digits), and `-`.
        It then passes this pattern to the pattern-based password generator to produce the final password.
        :return: The password and generation info.
        """
        pattern = "".join(self._rng.choices("/XXXxxx-999", k=28))
        pw = self._gen.generate_pattern_password(pattern)
        return {
            "extra_info": f"A pattern password with the pattern '{pattern}'.",
            "password": pw,
        }

    def complex(self) -> dict[str, str]:
        """
        Generate a complex password and return it with descriptive metadata.

        This method calls `generate_complex_password()` and returns a dictionary
        containing the password and a short description of its randomness and complexity.
        :return: The password and generation info.
        """
        pw = self._gen.generate_complex_password()
        return {"extra_info": "Entirely random, a complex password.", "password": pw}

    def mnemonic(self) -> dict[str, str]:
        """
        Generate a simple, memorable (mnemonic) password.

        This method generates a password using a common mnemonic pattern,
        typically involving an adjective and a noun (e.g., "BraveTiger42").
        It returns the generated password along with descriptive information
        about the words used.
        :return: The password and generation info.
        """
        pw, adj, noun = self._gen.generate_mnemonic_password(_return_info=True)
        return {
            "extra_info": f"A mnemonic password using the adj '{adj}' and the noun '{noun}'.",
            "password": pw,
        }

    def ratio(self) -> dict[str, str]:
        """
        Generate a ratio-based password with explanatory context.

        This method calls `generate_ratio_based_password_v3` using its default parameters
        and returns the password along with a brief description indicating that the
        password was generated entirely at random based on character-type ratios.
        :return: The password and generation info.
        """
        pw = self._gen.generate_ratio_based_password_v3()
        return {
            "extra_info": "Entirely random, a ratio based password.",
            "password": pw,
        }

    def words(self) -> dict[str, str]:
        """
        Generate a words-based password and return it with descriptive information.

        This method generates a password by calling `generate_words_based_password_v3`
        with `_return_info=True`, which returns both the password and the sentence used.
        The result includes a human-readable explanation of how the password was constructed.
        :return: The password and generation info.
        """
        pw, sentence = self._gen.generate_words_based_password_v3(_return_info=True)
        return {
            "extra_info": f"A words based password using the sentence '{sentence}_.",
            "password": pw,
        }

    def sentence(self) -> dict[str, str]:
        """
        Generate a sentence-based password with a randomized format and extra character.

        This method constructs a password by:
        - Randomly selecting a special character.
        - Randomizing the order of the password components: words, special characters, extra character, and numbers.
        - Calling the sentence-based password generator with the selected character and format.
        :return: The password and generation info.
        """
        spec_char = self._rng.choice("!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~")
        pattern = "".join(
            self._rng.sample(["{words}", "{special}", "{extra}", "{numbers}"], 4)
        )
        pw, sentence = self._gen.generate_sentence_based_password_v3(
            extra_char=spec_char, password_format=pattern, _return_info=True
        )
        return {
            "extra_info": f"A sentence based password using the sentence '{sentence}' "
            f"and the pattern '{pattern}'.",
            "password": pw,
        }

    def complex_pattern(self) -> dict[str, str]:
        """
        A pattern password with uppercase letters, lowercase letters, numbers, words, and symbols.
        :return: The password and generation info.
        """
        pattern = "".join(self._rng.choices("/XXXxxx-999nnWw!!!", k=28))
        pw = self._gen.generate_complex_pattern_password(pattern)
        return {
            "extra_info": f"A complex pattern password with the pattern '{pattern}'.",
            "password": pw,
        }

    def generate_secure_password(
        self, return_worst_case: bool = False
    ) -> dict[str, str]:
        """
        Generates always secure, always changing passwords that are designed to be human readable.
        The target worst time to crack is one century for each generated password.

        :param return_worst_case: Return the worst case estimate by zxcvbn
        :return:
        """
        rnd = int(self._exponential(0, 8, 0.9))
        result = [
            self.mnemonic,
            self.passphrase,
            self.pattern,
            self.complex_pattern,
            self.sentence,
            self.words,
            self.random,
            self.complex,
            self.ratio,
        ][rnd]()

        if return_worst_case:
            try:
                from zxcvbn import zxcvbn
            except ImportError:
                raise RuntimeError("zxcvbn is not installed")
            result["worst_case"] = zxcvbn(result["password"], max_length=800000)["crack_times_display"][
                "offline_fast_hashing_1e10_per_second"
            ]  # type: ignore

        return result



"""Fallback implementations"""

import hashlib
import hmac


def _hash_digest(data: bytes, algorithm: str) -> bytes:
    try:
        return hashlib.new(algorithm, data).digest()
    except ValueError as e:
        raise ValueError(
            f"Unsupported or unavailable hash algorithm: '{algorithm}'"
        ) from e


def hash_sha1(data: bytes) -> bytes:
    return _hash_digest(data, "sha1")


def hash_sha224(data: bytes) -> bytes:
    return _hash_digest(data, "sha224")


def hash_sha256(data: bytes) -> bytes:
    return _hash_digest(data, "sha256")


def hash_sha384(data: bytes) -> bytes:
    return _hash_digest(data, "sha384")


def hash_sha512(data: bytes) -> bytes:
    return _hash_digest(data, "sha512")


def hash_sha3_224(data: bytes) -> bytes:
    return _hash_digest(data, "sha3_224")


def hash_sha3_256(data: bytes) -> bytes:
    return _hash_digest(data, "sha3_256")


def hash_sha3_384(data: bytes) -> bytes:
    return _hash_digest(data, "sha3_384")


def hash_sha3_512(data: bytes) -> bytes:
    return _hash_digest(data, "sha3_512")


def hash_md2(data: bytes) -> bytes:
    """Might not work"""
    return _hash_digest(data, "md2")


def hash_md4(data: bytes) -> bytes:
    """Might not work"""
    return _hash_digest(data, "md4")


def hash_md5(data: bytes) -> bytes:
    return _hash_digest(data, "md5")


def hash_sm3(data: bytes) -> bytes:
    """Might not work"""
    return _hash_digest(data, "sm3")


def hash_ripemd160(data: bytes) -> bytes:
    """Might not work"""
    return _hash_digest(data, "ripemd160")


def hash_blake2b(data: bytes, digest_size: int = 64) -> bytes:
    if not 1 <= digest_size <= 64:
        raise ValueError("blake2b digest_size must be between 1 and 64 bytes")
    return hashlib.blake2b(data, digest_size=digest_size).digest()


def hash_blake2s(data: bytes, digest_size: int = 32) -> bytes:
    if not 1 <= digest_size <= 32:
        raise ValueError("blake2s digest_size must be between 1 and 32 bytes")
    return hashlib.blake2s(data, digest_size=digest_size).digest()


def hash_sha3_shake_128(data: bytes, length: int = 16) -> bytes:
    if length <= 0:
        raise ValueError("SHAKE128 output length must be positive")
    return hashlib.shake_128(data).digest(length)


def hash_sha3_shake_256(data: bytes, length: int = 32) -> bytes:
    if length <= 0:
        raise ValueError("SHAKE256 output length must be positive")
    return hashlib.shake_256(data).digest(length)


def hash_verify_sha1(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha1"))


def hash_verify_sha224(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha224"))


def hash_verify_sha256(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha256"))


def hash_verify_sha384(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha384"))


def hash_verify_sha512(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha512"))


def hash_verify_sha3_224(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha3_224"))


def hash_verify_sha3_256(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha3_256"))


def hash_verify_sha3_384(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha3_384"))


def hash_verify_sha3_512(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sha3_512"))


def hash_verify_md2(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "md2"))


def hash_verify_md4(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "md4"))


def hash_verify_md5(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "md5"))


def hash_verify_sm3(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "sm3"))


def hash_verify_ripemd160(data: bytes, expected_hash: bytes) -> bool:
    return hmac.compare_digest(expected_hash, _hash_digest(data, "ripemd160"))


def hash_verify_blake2b(data: bytes, expected_hash: bytes) -> bool:
    actual = hashlib.blake2b(data, digest_size=len(expected_hash)).digest()
    return hmac.compare_digest(expected_hash, actual)


def hash_verify_blake2s(data: bytes, expected_hash: bytes) -> bool:
    actual = hashlib.blake2s(data, digest_size=len(expected_hash)).digest()
    return hmac.compare_digest(expected_hash, actual)


def hash_verify_sha3_shake_128(data: bytes, expected_hash: bytes) -> bool:
    actual = hashlib.shake_128(data).digest(len(expected_hash))
    return hmac.compare_digest(expected_hash, actual)


def hash_verify_sha3_shake_256(data: bytes, expected_hash: bytes) -> bool:
    actual = hashlib.shake_256(data).digest(len(expected_hash))
    return hmac.compare_digest(expected_hash, actual)


def derive_pbkdf1(
    password: bytes, salt: bytes, length: int, iterations: int, hash_alg: str
) -> bytes:
    """
    Derive a key using PBKDF1 (legacy).
    """
    if not salt or len(salt) < 8:
        raise ValueError("Salt must be at least 8 bytes for PBKDF1")
    if iterations < 1:
        raise ValueError("PBKDF1 requires at least 1 iteration")

    try:
        digest_size = hashlib.new(hash_alg).digest_size
    except ValueError as e:
        raise ValueError(f"Unsupported hash algorithm: '{hash_alg}'") from e

    if length > digest_size:
        raise ValueError(
            f"PBKDF1 maximum output length is {digest_size} bytes for '{hash_alg}'"
        )

    dk = password + salt
    for _ in range(iterations):
        dk = hashlib.new(hash_alg, dk).digest()
    return dk[:length]


def derive_hkdf(
    password: bytes, salt: bytes, info: bytes, length: int, hash_alg: str
) -> bytes:
    """
    Derive a key using HKDF (RFC 5869).
    """
    try:
        hashlib.new(hash_alg)
    except ValueError as e:
        raise ValueError(f"Unsupported hash algorithm: '{hash_alg}'") from e

    def hmac_d(key: bytes, data: bytes) -> bytes:
        return hmac.new(key, data, hashlib.new(hash_alg).name).digest()

    prk = hmac_d(salt, password)

    okm, t, counter = b"", b"", 0
    while len(okm) < length:
        counter += 1
        t = hmac_d(prk, t + info + bytes([counter]))
        okm += t
    return okm[:length]


def derive_scrypt(
    password: bytes, salt: bytes, length: int, n: int, r: int, p: int
) -> bytes:
    """
    Derive a key using scrypt.
    """
    if len(salt) < 8:
        raise ValueError("Salt must be at least 8 bytes for scrypt")
    if not all(isinstance(x, int) and x > 0 for x in (n, r, p, length)):
        raise ValueError("scrypt parameters must be positive integers")
    # Enough maxmem to support the highest config of PasswordManager
    return hashlib.scrypt(password, salt=salt, n=n, r=r, p=p, dklen=length, maxmem=300*1024*1024)


def derive_pbkdf2hmac(
    password: bytes, salt: bytes, length: int, iterations: int, hash_alg: str
) -> bytes:
    """
    Derive a cryptographic key using PBKDF2.
    """
    if not isinstance(password, bytes):
        raise TypeError("password must be bytes")
    if not isinstance(salt, bytes):
        raise TypeError("salt must be bytes")
    if not isinstance(iterations, int) or iterations < 1:
        raise ValueError("iterations must be a positive integer")
    if not isinstance(length, int) or length < 1:
        raise ValueError("length must be a positive integer")

    try:
        return hashlib.pbkdf2_hmac(
            hash_alg.lower(), password, salt, iterations, dklen=length
        )
    except (ValueError, TypeError) as e:
        raise ValueError(f"PBKDF2 derivation failed: {e}")


def derive_concatkdf(
    shared_secret: bytes, length: int, otherinfo: bytes, hash_alg: str
) -> bytes:
    """
    Approximate ConcatKDF from NIST SP 800-56A using hashlib.
    """
    hash_len = hashlib.new(hash_alg).digest_size
    if length > 255 * hash_len:
        raise ValueError("Cannot derive more than 255 blocks of hash_len")

    output = b""
    counter = 1
    while len(output) < length:
        h = hashlib.new(hash_alg)
        h.update(counter.to_bytes(4, "big"))
        h.update(shared_secret)
        h.update(otherinfo)
        output += h.digest()
        counter += 1
    return output[:length]



from __future__ import annotations

import hmac
from typing import Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf import concatkdf, hkdf, pbkdf2, scrypt, x963kdf

try:
    from cryptography.hazmat.primitives.kdf.argon2 import Argon2id
except Exception:  # pragma: no cover - version dependent import
    Argon2id = None


# --- Hashes --- #


def _generic_hash(algorithm, data: bytes, digest_size: Optional[int] = None) -> bytes:
    if algorithm in (hashes.SHAKE128, hashes.SHAKE256, hashes.BLAKE2b, hashes.BLAKE2s):
        digest_size = digest_size or 32
        h = hashes.Hash(algorithm(digest_size), backend=default_backend())
    else:
        h = hashes.Hash(algorithm(), backend=default_backend())
    h.update(data)
    return h.finalize()


def _verify_hash(data: bytes, expected: bytes, algo_fn, *args, **kwargs) -> bool:
    return hmac.compare_digest(algo_fn(data, *args, **kwargs), expected)


def hash_sha1(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA1, data)


def hash_verify_sha1(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha1)


# SHA2

def hash_sha256(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA256, data)


def hash_verify_sha256(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha256)


def hash_sha384(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA384, data)


def hash_verify_sha384(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha384)


def hash_sha512(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA512, data)


def hash_verify_sha512(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha512)


def hash_sha224(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA224, data)


def hash_verify_sha224(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha224)


def hash_sha512_224(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA512_224, data)


def hash_verify_sha512_224(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha512_224)


def hash_sha512_256(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA512_256, data)


def hash_verify_sha512_256(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha512_256)


# SHA3

def hash_sha3_256(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA3_256, data)


def hash_verify_sha3_256(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha3_256)


def hash_sha3_512(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA3_512, data)


def hash_verify_sha3_512(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha3_512)


def hash_sha3_224(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA3_224, data)


def hash_verify_sha3_224(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha3_224)


def hash_sha3_384(data: bytes) -> bytes:
    return _generic_hash(hashes.SHA3_384, data)


def hash_verify_sha3_384(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sha3_384)


# SHAKE

def hash_sha3_shake_128(data: bytes, digest_size: int = 32) -> bytes:
    return _generic_hash(hashes.SHAKE128, data, digest_size)


def hash_verify_sha3_shake_128(data: bytes, expected: bytes, digest_size: int = 32) -> bool:
    return _verify_hash(data, expected, hash_sha3_shake_128, digest_size)


def hash_sha3_shake_256(data: bytes, digest_size: int = 32) -> bytes:
    return _generic_hash(hashes.SHAKE256, data, digest_size)


def hash_verify_sha3_shake_256(data: bytes, expected: bytes, digest_size: int = 32) -> bool:
    return _verify_hash(data, expected, hash_sha3_shake_256, digest_size)


# Others

def hash_blake2b(data: bytes, digest_size: int = 64) -> bytes:
    return _generic_hash(hashes.BLAKE2b, data, digest_size)


def hash_verify_blake2b(data: bytes, expected: bytes, digest_size: int = 64) -> bool:
    return _verify_hash(data, expected, hash_blake2b, digest_size)


def hash_blake2s(data: bytes, digest_size: int = 32) -> bytes:
    return _generic_hash(hashes.BLAKE2s, data, digest_size)


def hash_verify_blake2s(data: bytes, expected: bytes, digest_size: int = 32) -> bool:
    return _verify_hash(data, expected, hash_blake2s, digest_size)


def hash_md5(data: bytes) -> bytes:
    return _generic_hash(hashes.MD5, data)


def hash_verify_md5(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_md5)


def hash_sm3(data: bytes) -> bytes:
    return _generic_hash(hashes.SM3, data)


def hash_verify_sm3(data: bytes, expected: bytes) -> bool:
    return _verify_hash(data, expected, hash_sm3)


# --- KDFs --- #

algo_map = {
    "sha1": hashes.SHA1,
    "sha224": hashes.SHA224,
    "sha256": hashes.SHA256,
    "sha384": hashes.SHA384,
    "sha512": hashes.SHA512,
    "sha512_224": hashes.SHA512_224,
    "sha512_256": hashes.SHA512_256,
    "sha3_224": hashes.SHA3_224,
    "sha3_256": hashes.SHA3_256,
    "sha3_384": hashes.SHA3_384,
    "sha3_512": hashes.SHA3_512,
    "blake2b": hashes.BLAKE2b,
    "blake2s": hashes.BLAKE2s,
    "sm3": hashes.SM3,
    "md5": hashes.MD5,
}


def derive_pbkdf2hmac(password: bytes, salt: bytes, length: int, iterations: int, algorithm: str) -> bytes:
    algo_cls = algo_map.get(algorithm)
    if algo_cls is None:
        raise ValueError(f"Passed invalid/unsupported algorithm '{algorithm}'")
    return pbkdf2.PBKDF2HMAC(
        algorithm=algo_cls(),
        length=length,
        salt=salt,
        iterations=iterations,
        backend=default_backend(),
    ).derive(password)


def derive_scrypt(password: bytes, salt: bytes, length: int, n: int, r: int, p: int) -> bytes:
    return scrypt.Scrypt(salt=salt, length=length, n=n, r=r, p=p, backend=default_backend()).derive(password)


def derive_hkdf(password: bytes, salt: bytes, info: bytes, length: int, algorithm: str) -> bytes:
    algo_cls = algo_map.get(algorithm)
    if algo_cls is None:
        raise ValueError(f"Passed invalid/unsupported algorithm '{algorithm}'")
    return hkdf.HKDF(
        algorithm=algo_cls(),
        length=length,
        salt=salt,
        info=info,
        backend=default_backend(),
    ).derive(password)


def derive_x963(password: bytes, length: int, sharedinfo: bytes | None, algorithm: str) -> bytes:
    algo_cls = algo_map.get(algorithm)
    if algo_cls is None:
        raise ValueError(f"Passed invalid/unsupported algorithm '{algorithm}'")
    return x963kdf.X963KDF(
        algorithm=algo_cls(),
        length=length,
        sharedinfo=sharedinfo,
        backend=default_backend(),
    ).derive(password)


def derive_concatkdf(password: bytes, length: int, otherinfo: bytes | None, algorithm: str) -> bytes:
    algo_cls = algo_map.get(algorithm)
    if algo_cls is None:
        raise ValueError(f"Passed invalid/unsupported algorithm '{algorithm}'")
    return concatkdf.ConcatKDFHash(
        algorithm=algo_cls(),
        length=length,
        otherinfo=otherinfo,
        backend=default_backend(),
    ).derive(password)


def derive_argon2(
    password: bytes,
    salt: bytes,
    length: int,
    time_cost: int,
    memory_cost: int,
    parallelism: int,
    variant: str,
) -> bytes:
    if variant != "id":
        raise ValueError("cryptography Argon2 backend only supports variant='id'")
    if Argon2id is None:
        raise ValueError("Argon2id is unavailable in installed cryptography version")

    try:
        kdf = Argon2id(
            salt=salt,
            length=length,
            iterations=time_cost,
            lanes=parallelism,
            memory_cost=memory_cost,
        )
    except TypeError:
        # Compatibility fallback if constructor signature differs.
        kdf = Argon2id(
            salt=salt,
            length=length,
            time_cost=time_cost,
            parallelism=parallelism,
            memory_cost=memory_cost,
        )
    return kdf.derive(password)


def derive_kmac128(password: bytes, length: int, key: bytes = b"", customization: bytes = b"") -> bytes:
    kmac128 = getattr(hashes, "KMAC128", None)
    if kmac128 is None:
        raise ValueError("KMAC128 is unavailable in installed cryptography version")
    h = hashes.Hash(kmac128(key=key, length=length, customization_string=customization), backend=default_backend())
    h.update(password)
    return h.finalize()


def derive_kmac256(password: bytes, length: int, key: bytes = b"", customization: bytes = b"") -> bytes:
    kmac256 = getattr(hashes, "KMAC256", None)
    if kmac256 is None:
        raise ValueError("KMAC256 is unavailable in installed cryptography version")
    h = hashes.Hash(kmac256(key=key, length=length, customization_string=customization), backend=default_backend())
    h.update(password)
    return h.finalize()
